"""Module for base classes for the input models.

The philosophy of the input models is:

* Provide an easy way to input rules. The type hints are made to be human-friendly, for example, Literal instead of
  Enum.
* The .dump() method should fill out defaults and have shortcuts. For example, if the prefix is not provided for
  a class, then the prefix from the metadata is used. For views, if the class is not provided, it is assumed to
  be the same as the view.

The base classes are to make it easy to create the input models with default behavior. They are also used for
testing to ensure that input models correctly map to the verified rules models.
"""

import sys
from abc import ABC, abstractmethod
from dataclasses import Field, dataclass, fields, is_dataclass
from types import GenericAlias, UnionType
from typing import Any, Generic, TypeVar, Union, cast, get_args, get_origin, overload

import pandas as pd

from ._base_rules import BaseRules, SchemaModel

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

T_BaseRules = TypeVar("T_BaseRules", bound=BaseRules)
T_RuleModel = TypeVar("T_RuleModel", bound=SchemaModel)


@dataclass
class InputRules(Generic[T_BaseRules], ABC):
    """Input rules are raw data that is not yet validated."""

    @classmethod
    @abstractmethod
    def _get_verified_cls(cls) -> type[T_BaseRules]:
        raise NotImplementedError("This method should be implemented in the subclass.")

    @classmethod
    @overload
    def load(cls: "type[T_InputRules]", data: dict[str, Any]) -> "T_InputRules": ...

    @classmethod
    @overload
    def load(cls: "type[T_InputRules]", data: None) -> None: ...

    @classmethod
    def load(cls: "type[T_InputRules]", data: dict | None) -> "T_InputRules | None":
        if data is None:
            return None
        return cls._load(data)

    @classmethod
    def _type_by_field_name(cls) -> dict[str, type]:
        output: dict[str, type] = {}
        for field_ in fields(cls):
            type_ = field_.type
            if isinstance(type_, UnionType) or get_origin(type_) is Union:
                type_ = get_args(type_)[0]
            if isinstance(type_, str) and type_.startswith(cls.__name__):
                type_ = cls

            candidate: type
            if is_dataclass(type_):
                candidate = type_  # type: ignore[assignment]
            elif isinstance(type_, GenericAlias) and type_.__origin__ is list and is_dataclass(type_.__args__[0]):
                candidate = type_.__args__[0]  # type: ignore[assignment]

            # this handles prefixes
            elif isinstance(type_, GenericAlias) and type_.__origin__ is dict:
                candidate = type_  # type: ignore[assignment]
            else:
                continue

            output[field_.name] = candidate

        return output

    @classmethod
    def _load(cls, data: dict[str, Any]) -> Self:
        args: dict[str, Any] = {}
        field_type_by_name = cls._type_by_field_name()
        for field_name, field_ in cls._get_verified_cls().model_fields.items():
            field_type = field_type_by_name.get(field_name)
            if field_type is None:
                continue
            if field_name in data:
                value = data[field_name]
            elif field_.alias in data:
                value = data[field_.alias]
            else:
                continue

            # Handles the case where the field is a dataclass
            if hasattr(field_type, "_load"):
                if isinstance(value, dict):
                    args[field_name] = field_type._load(value)  # type: ignore[attr-defined]
                elif isinstance(value, list) and value and isinstance(value[0], dict):
                    args[field_name] = [field_type._load(item) for item in value]  # type: ignore[attr-defined]
            # Handles the case where the field holds non-dataclass values, e.g. a prefixes dict
            else:
                args[field_name] = value

        return cls(**args)

    def _dataclass_fields(self) -> list[Field]:
        return list(fields(self))

    def as_verified_rules(self) -> T_BaseRules:
        cls_ = self._get_verified_cls()
        return cls_.model_validate(self.dump())

    def dump(self) -> dict[str, Any]:
        output: dict[str, Any] = {}
        for field_ in self._dataclass_fields():
            value = getattr(self, field_.name)
            if value is None:
                continue
            if hasattr(value, "dump"):
                output[field_.name] = value.dump()
            elif isinstance(value, list) and value and hasattr(value[0], "dump"):
                output[field_.name] = [item.dump() for item in value]
        return output


T_InputRules = TypeVar("T_InputRules", bound=InputRules)


@dataclass
class InputComponent(ABC, Generic[T_RuleModel]):
    @classmethod
    @abstractmethod
    def _get_verified_cls(cls) -> type[T_RuleModel]:
        raise NotImplementedError("This method should be implemented in the subclass.")

    @classmethod
    @overload
    def load(cls, data: None) -> None: ...

    @classmethod
    @overload
    def load(cls, data: dict[str, Any]) -> Self: ...

    @classmethod
    @overload
    def load(cls, data: list[dict[str, Any]]) -> list[Self]: ...

    @classmethod
    def load(cls, data: dict[str, Any] | list[dict[str, Any]] | None) -> Self | list[Self] | None:
        if data is None:
            return None
        if isinstance(data, list) or (isinstance(data, dict) and isinstance(data.get("data"), list)):
            items = cast(list[dict[str, Any]], data.get("data") if isinstance(data, dict) else data)
            return [loaded for item in items if (loaded := cls.load(item)) is not None]
        return cls._load(data)

    @classmethod
    def _load(cls, data: dict[str, Any]) -> Self:
        args: dict[str, Any] = {}
        for field_name, field_ in cls._get_verified_cls().model_fields.items():  # type: ignore[attr-defined]
            if field_.exclude:
                continue
            if field_name in data:
                args[field_name] = data[field_name]
            elif field_.alias in data:
                args[field_name] = data[field_.alias]
        return cls(**args)

    def dump(self, **kwargs) -> dict[str, Any]:
        return {
            field_.alias or name: getattr(self, name)
            for name, field_ in self._get_verified_cls().model_fields.items()
            if not field_.exclude
        }

    def _repr_html_(self) -> str:
        return pd.DataFrame([self.dump()])._repr_html_()  # type: ignore[operator]



from collections import defaultdict
from collections.abc import Iterable, Sequence
from graphlib import CycleError, TopologicalSorter
from typing import TYPE_CHECKING

from cognite.client import data_modeling as dm

from cognite.neat._client.data_classes.data_modeling import ContainerApplyDict, SpaceApplyDict, ViewApplyDict
from cognite.neat._client.data_classes.schema import DMSSchema
from cognite.neat._constants import is_hierarchy_property
from cognite.neat._issues.errors import NeatValueError

if TYPE_CHECKING:
    from cognite.neat._client._api_client import NeatClient


class SchemaAPI:
    def __init__(self, client: "NeatClient") -> None:
        self._client = client

    def retrieve(
        self,
        view_ids: Sequence[dm.ViewId],
        container_ids: Sequence[dm.ContainerId],
        include_ancestors: bool = True,
        include_connections: bool = True,
        data_model_id: dm.DataModelId | None = None,
    ) -> DMSSchema:
        data_model_id = data_model_id or dm.DataModelId("NEAT_LOOKUP", "NEAT_LOOKUP", "NEAT_LOOKUP")
        if data_model_id.version is None:
            raise NeatValueError("Data model version must be specified")
        read_views = self._client.loaders.views.retrieve(
            list(view_ids),
            format="read",
            include_connected=include_connections,
            include_ancestor=include_ancestors,
        )
        views = ViewApplyDict([self._client.loaders.views.as_write(view) for view in read_views])

        container_set = set(container_ids) | {
            container for view in read_views for container in view.referenced_containers()
        }
        containers = self._client.loaders.containers.retrieve(list(container_set))

        return DMSSchema(
            data_model=dm.DataModelApply(
                space=data_model_id.space,
                external_id=data_model_id.external_id,
                version=data_model_id.version,
                views=list(views.keys()),
            ),
            views=views,
            containers=ContainerApplyDict(containers.as_write()),
        )

    def retrieve_data_model_id(self, data_model_id: dm.DataModelIdentifier) -> DMSSchema:
        data_models = self._client.data_modeling.data_models.retrieve(data_model_id, inline_views=True)
        if len(data_models) == 0:
            raise NeatValueError(f"Data model {data_model_id} not found")
        data_model = data_models.latest_version()
        return self.retrieve_data_model(data_model)

    def retrieve_data_model(
        self,
        data_model: dm.DataModel[dm.View],
    ) -> DMSSchema:
        """Create a schema from a data model.

        If a reference model is provided, the schema will include a reference schema. To determine which views,
        and containers to put in the reference schema, the following rule is applied:

            If a view or container space is different from the data model space,
            it will be included in the reference schema.*

        *One exception to this rule is if a view is directly referenced by the data model. In this case, the view will
        be included in the data model schema, even if the space is different.

        Args:
            data_model: The data model to create the schema from.

        Returns:
            DMSSchema: The schema created from the data model.
        """
        data_model_views = dm.ViewList(data_model.views)
        data_model_write = data_model.as_write()
        data_model_write.views = list(data_model_views.as_ids())

        container_ids = data_model_views.referenced_containers()
        containers = self._client.loaders.containers.retrieve(list(container_ids), include_connected=True)

        space_ids = [data_model.space]
        space_read = self._client.loaders.spaces.retrieve(space_ids)
        if len(space_read) != len(space_ids):
            raise NeatValueError(f"Space(s) {space_read} not found")
        space_write = space_read.as_write()

        existing_view_ids = set(data_model_views.as_ids())
        views_with_referenced = self._client.loaders.views.retrieve(
            list(existing_view_ids), include_connected=True, include_ancestor=True
        )

        # Converting views from read to write format requires to account for parents (implements)
        # as the read format contains all properties from all parents, while the write formate should not contain
        # properties from any parents.
        # The ViewLoader as_write method looks up parents and remove properties from them.
        view_write = ViewApplyDict([self._client.loaders.views.as_write(view) for view in views_with_referenced])

        container_write = ContainerApplyDict(containers.as_write())
        user_space = data_model.space
        return DMSSchema(
            spaces=SpaceApplyDict([s for s in space_write if s.space == user_space]),
            data_model=data_model_write,
            views=view_write,
            containers=container_write,
        )

    @staticmethod
    def get_hierarchical_properties(
        views: Iterable[dm.View | dm.ViewApply],
    ) -> dict[dm.ViewId, set[str]]:
        """Sorts the views by container constraints."""
        hierarchical_properties_by_view_id: dict[dm.ViewId, set[str]] = defaultdict(set)
        for view in views:
            for prop_id, prop in (view.properties or {}).items():
                if not isinstance(prop, dm.MappedProperty | dm.MappedPropertyApply):
                    continue
                if is_hierarchy_property(prop.container, prop.container_property_identifier):
                    hierarchical_properties_by_view_id[view.as_id()].add(prop_id)
        return hierarchical_properties_by_view_id

    @staticmethod
    def get_view_order_by_direct_relation_constraints(views: Iterable[dm.View]) -> list[dm.ViewId]:
        """Sorts the views by container constraints."""
        view_sequence = list(views)
        view_ids_by_container: dict[dm.ContainerId, set[dm.ViewId]] = defaultdict(set)
        for view in view_sequence:
            for container_id in view.referenced_containers():
                view_ids_by_container[container_id].add(view.as_id())

        view_by_dependency: dict[dm.ViewId, set[dm.ViewId]] = {}
        for view in view_sequence:
            view_id = view.as_id()
            view_by_dependency[view_id] = set()
            for prop in view.properties.values():
                if (
                    isinstance(prop, dm.MappedProperty)
                    and isinstance(prop.type, dm.DirectRelation)
                    and prop.type.container
                ):
                    view_by_dependency[view_id].update(view_ids_by_container[prop.type.container])
        try:
            return list(TopologicalSorter(view_by_dependency).static_order())
        except CycleError as e:
            raise NeatValueError("Cycle in direct relation constraints") from e



"""These are user modeling warnings. These warnings will not cause the resulting model
to be invalid, but will likely lead to suboptimal performance, unnecessary complexity,
or other issues."""
# These warnings are usually only used once in the code base.

from dataclasses import dataclass

from cognite.client.data_classes.data_modeling import ContainerId, ViewId

from cognite.neat._constants import DMS_CONTAINER_PROPERTY_SIZE_LIMIT

from ._models import UserModelingWarning

__all__ = [
    "ContainerPropertyLimitWarning",
    "DirectRelationMissingSourceWarning",
    "EmptyContainerWarning",
    "HasDataFilterOnNoPropertiesViewWarning",
    "HasDataFilterOnViewWithReferencesWarning",
    "NodeTypeFilterOnParentViewWarning",
    "NotNeatSupportedFilterWarning",
    "ParentInDifferentSpaceWarning",
    "ViewsAndDataModelNotInSameSpaceWarning",
]


@dataclass(unsafe_hash=True)
class ViewsAndDataModelNotInSameSpaceWarning(UserModelingWarning):
    """The data model is in {data_model_space}, while views are in {views_spaces} space(s).
    This is strongly discouraged as it can lead to confusion and unnecessary complexity.
    """

    fix = "Ensure that views and data model are in the same space and have same version"
    data_model_space: str
    views_spaces: str


@dataclass(unsafe_hash=True)
class DirectRelationMissingSourceWarning(UserModelingWarning):
    """The view {view_id}.{prop_name} is a direct relation without a source.
    Direct relations in views should point to a single other view, if not, you end up
    with a more complex model than necessary."""

    fix = "Create the source view"

    view_id: ViewId
    prop_name: str


@dataclass(unsafe_hash=True)
class ParentInDifferentSpaceWarning(UserModelingWarning):
    """The view {view_id} has multiple parents in different spaces.
    Neat recommends maximum one implementation of a view from another space."""

    fix = "Ensure all parents of the view are in the same space"

    view_id: ViewId


@dataclass(unsafe_hash=True)
class EmptyContainerWarning(UserModelingWarning):
    """Container {container_id} is empty and will be skipped.
    The container does not have any properties."""

    fix = "Add properties to the container or remove the container"

    container_id: ContainerId


@dataclass(unsafe_hash=True)
class HasDataFilterOnNoPropertiesViewWarning(UserModelingWarning):
    """Cannot set hasData filter on view {view_id}.
    The view does not have properties in any containers.
    Use a node type filter instead."""

    fix = "Use a node type filter instead"

    view_id: ViewId


@dataclass(unsafe_hash=True)
class NodeTypeFilterOnParentViewWarning(UserModelingWarning):
    """Setting a node type filter on parent view {view_id}.
    This is not recommended as parent views are typically used for multiple types of nodes."""

    fix = "Use a HasData filter instead"

    view_id: ViewId


@dataclass(unsafe_hash=True)
class HasDataFilterOnViewWithReferencesWarning(UserModelingWarning):
    """Setting a hasData filter on view {view_id} which references other views {references}.
    This is not recommended as it will lead to no nodes being returned when querying the solution view.
    Use a NodeType filter instead."""

    fix = "Use a NodeType filter instead"

    view_id: ViewId
    references: frozenset[ViewId]


@dataclass(unsafe_hash=True)
class ContainerPropertyLimitWarning(UserModelingWarning):
    """The number of properties in the {container_id} view is {count} which
    is more than the API limit {limit} properties.
    This can lead to performance issues.
    Reduce the number of properties in the view."""

    fix = "Reduce the number of properties in the view"

    container_id: ContainerId
    count: int
    limit: int = DMS_CONTAINER_PROPERTY_SIZE_LIMIT


@dataclass(unsafe_hash=True)
class NotNeatSupportedFilterWarning(UserModelingWarning):
    """The view {view_id} uses a non-standard filter.
    This will not be validated by Neat, and is thus not recommended.
    If you can use a HasData or NoteType filter."""

    fix = "Use a HasData or NoteType filter instead"

    view_id: ViewId



from collections import Counter
from collections.abc import Iterable, Sequence
from typing import TypeVar

from cognite.neat._config import GLOBAL_CONFIG
from cognite.neat._constants import IN_PYODIDE

T_Element = TypeVar("T_Element")


def remove_none_elements_from_set(s: set[T_Element]) -> set[T_Element]:
    return {x for x in s if x is not None}


def most_occurring_element(list_of_elements: list[T_Element]) -> T_Element:
    counts = Counter(list_of_elements)
    return counts.most_common(1)[0][0]


def chunker(sequence: Sequence[T_Element], chunk_size: int) -> Iterable[Sequence[T_Element]]:
    for i in range(0, len(sequence), chunk_size):
        yield sequence[i : i + chunk_size]


def remove_list_elements(input_list: list, elements_to_remove: list) -> list:
    return [element for element in input_list if element not in elements_to_remove]


def iterate_progress_bar(iterable: Iterable[T_Element], total: int, description: str) -> Iterable[T_Element]:
    if IN_PYODIDE or GLOBAL_CONFIG.progress_bar in ("infer", "tqdm"):
        try:
            from tqdm import tqdm
        except ModuleNotFoundError:
            return iterable
        return tqdm(iterable, total=total, desc=description)

    elif GLOBAL_CONFIG.progress_bar == "tqdm-notebook":
        try:
            from tqdm.notebook import tqdm as tqdm_notebook
        except ModuleNotFoundError:
            return iterable
        return tqdm_notebook(iterable, total=total, desc=description)
    elif GLOBAL_CONFIG.progress_bar in ("infer", "rich"):
        # Progress bar from rich requires multi-threading, which is not supported in Pyodide
        try:
            from rich.progress import track
        except ModuleNotFoundError:
            return iterable

        return track(
            iterable,
            total=total,
            description=description,
        )
    elif GLOBAL_CONFIG.progress_bar is None:
        return iterable
    else:
        raise ValueError(f"Unsupported progress bar type: {GLOBAL_CONFIG.progress_bar}")


def iterate_progress_bar_if_above_config_threshold(
    iterable: Iterable[T_Element], total: int, description: str
) -> Iterable[T_Element]:
    if GLOBAL_CONFIG.use_iterate_bar_threshold and total > GLOBAL_CONFIG.use_iterate_bar_threshold:
        return iterate_progress_bar(iterable, total, description)
    return iterable



from collections import defaultdict
from collections.abc import Collection, Iterable, Sequence
from datetime import datetime
from pathlib import Path
from typing import Literal, cast

from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling import DataModelId, DataModelIdentifier
from cognite.client.data_classes.data_modeling.containers import BTreeIndex, InvertedIndex
from cognite.client.data_classes.data_modeling.data_types import Enum as DMSEnum
from cognite.client.data_classes.data_modeling.data_types import ListablePropertyType, PropertyTypeWithUnit
from cognite.client.data_classes.data_modeling.views import (
    MultiEdgeConnectionApply,
    MultiReverseDirectRelationApply,
    SingleEdgeConnectionApply,
    SingleReverseDirectRelationApply,
    ViewPropertyApply,
)
from cognite.client.utils import ms_to_datetime

from cognite.neat._client import NeatClient
from cognite.neat._issues import IssueList, MultiValueError, NeatIssue, catch_issues
from cognite.neat._issues.errors import (
    FileTypeUnexpectedError,
    NeatValueError,
    ResourceMissingIdentifierError,
    ResourceRetrievalError,
)
from cognite.neat._issues.warnings import (
    MissingCogniteClientWarning,
    PropertyNotFoundWarning,
    PropertyTypeNotSupportedWarning,
    ResourceNotFoundWarning,
    ResourcesDuplicatedWarning,
    ResourceUnknownWarning,
)
from cognite.neat._rules._shared import ReadRules
from cognite.neat._rules.importers._base import BaseImporter
from cognite.neat._rules.models import (
    DMSInputRules,
    DMSSchema,
)
from cognite.neat._rules.models.data_types import DataType, Enum
from cognite.neat._rules.models.dms import (
    DMSInputContainer,
    DMSInputEnum,
    DMSInputMetadata,
    DMSInputNode,
    DMSInputProperty,
    DMSInputView,
)
from cognite.neat._rules.models.entities import (
    ClassEntity,
    ContainerEntity,
    DMSNodeEntity,
    DMSUnknownEntity,
    EdgeEntity,
    ReverseConnectionEntity,
    ViewEntity,
)


class DMSImporter(BaseImporter[DMSInputRules]):
    """Imports a Data Model from Cognite Data Fusion.

    Args:
        schema: The schema containing the data model.
        read_issues: A list of issues that occurred during the import.
        metadata: Metadata for the data model.

    """

    def __init__(
        self,
        schema: DMSSchema,
        read_issues: Sequence[NeatIssue] | None = None,
        metadata: DMSInputMetadata | None = None,
        referenced_containers: Iterable[dm.ContainerApply] | None = None,
    ):
        self.schema = schema
        self.metadata = metadata
        self.issue_list = IssueList(read_issues)
        self._all_containers_by_id = schema.containers.copy()
        self._all_views_by_id = schema.views.copy()
        if referenced_containers is not None:
            for container in referenced_containers:
                if container.as_id() in self._all_containers_by_id:
                    continue
                self._all_containers_by_id[container.as_id()] = container

    @property
    def description(self) -> str:
        if self.schema.data_model is not None:
            identifier = f"{self.schema.data_model.as_id().as_tuple()!s}"
        else:
            identifier = "Unknown"
        return f"DMS Data model {identifier} read as unverified data model"

    @classmethod
    def from_data_model_id(
        cls,
        client: NeatClient,
        data_model_id: DataModelIdentifier,
    ) -> "DMSImporter":
        """Create a DMSImporter ready to convert the given data model to rules.

        Args:
            client: Instantiated CogniteClient to retrieve data model.
            data_model_id: Data Model to retrieve.

        Returns:
            DMSImporter: DMSImporter instance
        """

        data_model_ids = [data_model_id]
        data_models = client.data_modeling.data_models.retrieve(data_model_ids, inline_views=True)

        retrieved_models = cls._find_model_in_list(data_models, data_model_id)
        if len(retrieved_models) == 0:
            return cls(
                DMSSchema(),
                [
                    ResourceRetrievalError(
                        dm.DataModelId.load(data_model_id),  # type: ignore[arg-type]
                        "data model",
                        "Data Model is missing in CDF",
                    )
                ],
            )
        return cls.from_data_model(client, retrieved_models.latest_version())

    @classmethod
    def from_data_model(cls, client: NeatClient, model: dm.DataModel[dm.View]) -> "DMSImporter":
        with catch_issues() as issue_list:
            schema = client.schema.retrieve_data_model(model)

        if issue_list.has_errors:
            return cls(DMSSchema(), issue_list)

        metadata = cls._create_metadata_from_model(model)

        return cls(
            schema,
            issue_list,
            metadata,
            referenced_containers=cls._lookup_referenced_containers(schema, issue_list, client),
        )

    @classmethod
    def _find_model_in_list(
        cls, data_models: dm.DataModelList[dm.View], model_id: DataModelIdentifier
    ) -> dm.DataModelList[dm.View]:
        identifier = DataModelId.load(model_id)
        return dm.DataModelList[dm.View](
            [
                model
                for model in data_models
                if (model.space, model.external_id) == (identifier.space, identifier.external_id)
            ]
        )

    @classmethod
    def _create_metadata_from_model(
        cls,
        model: dm.DataModel[dm.View] | dm.DataModelApply,
    ) -> DMSInputMetadata:
        description, creator = DMSInputMetadata._get_description_and_creator(model.description)

        if isinstance(model, dm.DataModel):
            created = ms_to_datetime(model.created_time)
            updated = ms_to_datetime(model.last_updated_time)
        else:
            now = datetime.now().replace(microsecond=0)
            created = now
            updated = now
        return DMSInputMetadata(
            space=model.space,
            external_id=model.external_id,
            name=model.name or model.external_id,
            version=model.version or "0.1.0",
            updated=updated,
            created=created,
            creator=",".join(creator),
            description=description,
        )

    @classmethod
    def from_directory(cls, directory: str | Path, client: NeatClient | None = None) -> "DMSImporter":
        with catch_issues() as issue_list:
            schema = DMSSchema.from_directory(directory)
        # If there were errors during the import, the to_rules will raise them.
        return cls(
            schema, issue_list, referenced_containers=cls._lookup_referenced_containers(schema, issue_list, client)
        )

    @classmethod
    def from_zip_file(cls, zip_file: str | Path, client: NeatClient | None = None) -> "DMSImporter":
        if Path(zip_file).suffix != ".zip":
            return cls(
                DMSSchema(),
                [FileTypeUnexpectedError(Path(zip_file), frozenset([".zip"]))],
            )
        with catch_issues() as issue_list:
            schema = DMSSchema.from_zip(zip_file)
        return cls(
            schema, issue_list, referenced_containers=cls._lookup_referenced_containers(schema, issue_list, client)
        )

    @classmethod
    def _lookup_referenced_containers(
        cls, schema: DMSSchema, issue_list: IssueList, client: NeatClient | None = None
    ) -> Iterable[dm.ContainerApply]:
        ref_containers = schema.externally_referenced_containers()
        if not ref_containers:
            return []
        elif client is None:
            id_ = ""
            if schema.data_model:
                id_ = f" {schema.data_model.as_id()!r}"
            issue_list.append(MissingCogniteClientWarning(f"importing full DMS model{id_}"))
            return []
        return client.loaders.containers.retrieve(list(ref_containers), format="write")

    @classmethod
    def from_path(cls, path: Path, client: NeatClient | None = None) -> "DMSImporter":
        if path.is_file():
            return cls.from_zip_file(path, client)
        elif path.is_dir():
            return cls.from_directory(path, client)
        else:
            raise NeatValueError(f"Unsupported YAML format: {format}")

    def to_rules(self) -> ReadRules[DMSInputRules]:
        if self.issue_list.has_errors:
            # In case there were errors during the import, the to_rules method will return None
            self.issue_list.trigger_warnings()
            raise MultiValueError(self.issue_list.errors)

        if not self.schema.data_model:
            self.issue_list.append(ResourceMissingIdentifierError("data model", type(self.schema).__name__))
            self.issue_list.trigger_warnings()
            raise MultiValueError(self.issue_list.errors)

        model = self.schema.data_model

        user_rules = self._create_rule_components(model, self.schema, self.metadata)

        self.issue_list.trigger_warnings()
        if self.issue_list.has_errors:
            raise MultiValueError(self.issue_list.errors)
        return ReadRules(user_rules, {})

    def _create_rule_components(
        self,
        data_model: dm.DataModelApply,
        schema: DMSSchema,
        metadata: DMSInputMetadata | None = None,
    ) -> DMSInputRules:
        enum_by_container_property = self._create_enum_collections(self._all_containers_by_id.values())
        enum_collection_by_container_property = {
            key: enum_list[0].collection for key, enum_list in enum_by_container_property.items() if enum_list
        }

        properties: list[DMSInputProperty] = []
        for view_id, view in schema.views.items():
            view_entity = ViewEntity.from_id(view_id)
            for prop_id, prop in (view.properties or {}).items():
                dms_property = self._create_dms_property(
                    prop_id, prop, view_entity, enum_collection_by_container_property
                )
                if dms_property is not None:
                    properties.append(dms_property)

        data_model_view_ids: set[dm.ViewId] = {
            view.as_id() if isinstance(view, dm.View | dm.ViewApply) else view for view in data_model.views or []
        }

        metadata = metadata or DMSInputMetadata.from_data_model(data_model)

        return DMSInputRules(
            metadata=metadata,
            properties=properties,
            containers=[DMSInputContainer.from_container(container) for container in schema.containers.values()],
            views=[
                DMSInputView.from_view(view, in_model=view_id in data_model_view_ids)
                for view_id, view in schema.views.items()
            ],
            nodes=[DMSInputNode.from_node_type(node_type) for node_type in schema.node_types.values()],
            enum=[enum for enum_list in enum_by_container_property.values() for enum in enum_list] or None,
        )

    def _create_dms_property(
        self,
        prop_id: str,
        prop: ViewPropertyApply,
        view_entity: ViewEntity,
        enum_collection_by_container_property: dict[tuple[dm.ContainerId, str], str],
    ) -> DMSInputProperty | None:
        if isinstance(prop, dm.MappedPropertyApply) and prop.container not in self._all_containers_by_id:
            self.issue_list.append(
                ResourceNotFoundWarning[dm.ContainerId, dm.PropertyId](
                    dm.ContainerId.load(prop.container),
                    "container",
                    view_entity.to_property_id(prop_id),
                    "view property",
                )
            )
            return None
        if (
            isinstance(prop, dm.MappedPropertyApply)
            and prop.container_property_identifier not in self._all_containers_by_id[prop.container].properties
        ):
            self.issue_list.append(
                PropertyNotFoundWarning(prop.container, "container", prop_id, view_entity.as_id(), "view"),
            )
            return None
        if not isinstance(
            prop,
            dm.MappedPropertyApply
            | SingleEdgeConnectionApply
            | MultiEdgeConnectionApply
            | SingleReverseDirectRelationApply
            | MultiReverseDirectRelationApply,
        ):
            self.issue_list.append(
                PropertyTypeNotSupportedWarning[dm.ViewId](view_entity.as_id(), "view", prop_id, type(prop).__name__)
            )
            return None

        value_type = self._get_value_type(prop, view_entity, prop_id, enum_collection_by_container_property)
        if value_type is None:
            return None

        return DMSInputProperty(
            description=prop.description,
            name=prop.name,
            connection=self._get_connection_type(prop),
            value_type=str(value_type),
            is_list=self._get_is_list(prop),
            nullable=self._get_nullable(prop),
            immutable=self._get_immutable(prop),
            default=self._get_default(prop),
            container=(
                str(ContainerEntity.from_id(prop.container)) if isinstance(prop, dm.MappedPropertyApply) else None
            ),
            container_property=(
                prop.container_property_identifier if isinstance(prop, dm.MappedPropertyApply) else None
            ),
            view=str(view_entity),
            view_property=prop_id,
            index=self._get_index(prop, prop_id),
            constraint=self._get_constraint(prop, prop_id),
        )

    def _container_prop_unsafe(self, prop: dm.MappedPropertyApply) -> dm.ContainerProperty:
        """This method assumes you have already checked that the container with property exists."""
        return self._all_containers_by_id[prop.container].properties[prop.container_property_identifier]

    def _get_connection_type(
        self, prop: ViewPropertyApply
    ) -> Literal["direct"] | ReverseConnectionEntity | EdgeEntity | None:
        if isinstance(prop, SingleEdgeConnectionApply | MultiEdgeConnectionApply):
            properties = ViewEntity.from_id(prop.edge_source) if prop.edge_source is not None else None
            return EdgeEntity(
                properties=properties, type=DMSNodeEntity.from_reference(prop.type), direction=prop.direction
            )
        elif isinstance(prop, SingleReverseDirectRelationApply | MultiReverseDirectRelationApply):
            return ReverseConnectionEntity(property=prop.through.property)
        elif isinstance(prop, dm.MappedPropertyApply) and isinstance(
            self._container_prop_unsafe(prop).type, dm.DirectRelation
        ):
            return "direct"
        else:
            return None

    def _get_value_type(
        self,
        prop: ViewPropertyApply,
        view_entity: ViewEntity,
        prop_id: str,
        enum_collection_by_container_property: dict[tuple[dm.ContainerId, str], str],
    ) -> DataType | ViewEntity | DMSUnknownEntity | None:
        if isinstance(
            prop,
            SingleEdgeConnectionApply
            | MultiEdgeConnectionApply
            | SingleReverseDirectRelationApply
            | MultiReverseDirectRelationApply,
        ):
            return ViewEntity.from_id(prop.source)
        elif isinstance(prop, dm.MappedPropertyApply):
            container_prop = self._container_prop_unsafe(cast(dm.MappedPropertyApply, prop))
            if isinstance(container_prop.type, dm.DirectRelation):
                if prop.source is None:
                    return DMSUnknownEntity()
                elif prop.source not in self._all_views_by_id:
                    self.issue_list.append(ResourceUnknownWarning(prop.source, "view", view_entity.as_id(), "view"))
                    return ViewEntity.from_id(prop.source)
                else:
                    return ViewEntity.from_id(prop.source)
            elif isinstance(container_prop.type, PropertyTypeWithUnit) and container_prop.type.unit:
                return DataType.load(f"{container_prop.type._type}(unit={container_prop.type.unit.external_id})")
            elif isinstance(container_prop.type, DMSEnum):
                collection = enum_collection_by_container_property.get(
                    (prop.container, prop.container_property_identifier)
                )
                if collection is None:
                    # This should never happen
                    raise ValueError(
                        f"BUG in Neat: Enum for {prop.container}.{prop.container_property_identifier} not found."
                    )

                return Enum(collection=ClassEntity(suffix=collection), unknownValue=container_prop.type.unknown_value)
            else:
                return DataType.load(container_prop.type._type)
        else:
            self.issue_list.append(
                PropertyTypeNotSupportedWarning[dm.ViewId](view_entity.as_id(), "view", prop_id, type(prop).__name__)
            )
            return None

    def _get_nullable(self, prop: ViewPropertyApply) -> bool | None:
        if isinstance(prop, dm.MappedPropertyApply):
            return self._container_prop_unsafe(prop).nullable
        else:
            return None

    def _get_immutable(self, prop: ViewPropertyApply) -> bool | None:
        if isinstance(prop, dm.MappedPropertyApply):
            return self._container_prop_unsafe(prop).immutable
        else:
            return None

    def _get_is_list(self, prop: ViewPropertyApply) -> bool | None:
        if isinstance(prop, dm.MappedPropertyApply):
            prop_type = self._container_prop_unsafe(prop).type
            return isinstance(prop_type, ListablePropertyType) and prop_type.is_list
        elif isinstance(prop, MultiEdgeConnectionApply | MultiReverseDirectRelationApply):
            return True
        elif isinstance(prop, SingleEdgeConnectionApply | SingleReverseDirectRelationApply):
            return False
        else:
            return None

    def _get_default(self, prop: ViewPropertyApply) -> str | None:
        if isinstance(prop, dm.MappedPropertyApply):
            default = self._container_prop_unsafe(prop).default_value
            if default is not None:
                return str(default)
        return None

    def _get_index(self, prop: ViewPropertyApply, prop_id) -> list[str] | None:
        if not isinstance(prop, dm.MappedPropertyApply):
            return None
        container = self._all_containers_by_id[prop.container]
        index: list[str] = []
        for index_name, index_obj in (container.indexes or {}).items():
            if isinstance(index_obj, BTreeIndex | InvertedIndex) and prop_id in index_obj.properties:
                index.append(index_name)
        return index or None

    def _get_constraint(self, prop: ViewPropertyApply, prop_id: str) -> list[str] | None:
        if not isinstance(prop, dm.MappedPropertyApply):
            return None
        container = self._all_containers_by_id[prop.container]
        unique_constraints: list[str] = []
        for constraint_name, constraint_obj in (container.constraints or {}).items():
            if isinstance(constraint_obj, dm.RequiresConstraint):
                # This is handled in the .from_container method of DMSContainer
                continue
            elif isinstance(constraint_obj, dm.UniquenessConstraint) and prop_id in constraint_obj.properties:
                unique_constraints.append(constraint_name)
            elif isinstance(constraint_obj, dm.UniquenessConstraint):
                # This does not apply to this property
                continue
            else:
                self.issue_list.append(
                    PropertyTypeNotSupportedWarning[dm.ContainerId](
                        prop.container, "container", prop_id, type(constraint_obj).__name__
                    )
                )
        return unique_constraints or None

    def _find_reverse_edge(
        self, prop_id: str, prop: SingleEdgeConnectionApply | MultiEdgeConnectionApply, view_id: dm.ViewId
    ) -> str | None:
        if prop.source not in self._all_views_by_id:
            return None
        view = self._all_views_by_id[prop.source]
        candidates = []
        for prop_name, reverse_prop in (view.properties or {}).items():
            if isinstance(reverse_prop, SingleEdgeConnectionApply | MultiEdgeConnectionApply):
                if (
                    reverse_prop.type == prop.type
                    and reverse_prop.source == view_id
                    and reverse_prop.direction != prop.direction
                ):
                    candidates.append(prop_name)
        if len(candidates) == 0:
            self.issue_list.append(
                PropertyNotFoundWarning(
                    prop.source,
                    "view property",
                    f"reverse edge of {prop_id}",
                    dm.PropertyId(view_id, prop_id),
                    "view property",
                )
            )
            return None
        if len(candidates) > 1:
            self.issue_list.append(
                ResourcesDuplicatedWarning(
                    frozenset(dm.PropertyId(view.as_id(), candidate) for candidate in candidates),
                    "view property",
                    default_action="Multiple reverse edges found for "
                    f"{dm.PropertyId(view_id, prop_id)!r}. Will use {candidates[0]}",
                )
            )

        return candidates[0]

    @staticmethod
    def _create_enum_collections(
        containers: Collection[dm.ContainerApply],
    ) -> dict[tuple[dm.ContainerId, str], list[DMSInputEnum]]:
        enum_by_container_property: dict[tuple[dm.ContainerId, str], list[DMSInputEnum]] = defaultdict(list)

        is_external_id_unique = len({container.external_id for container in containers}) == len(containers)

        for container in containers:
            container_id = container.as_id()
            for prop_id, prop in container.properties.items():
                if not isinstance(prop.type, DMSEnum):
                    continue
                if is_external_id_unique:
                    collection = f"{container.external_id}.{prop_id}"
                else:
                    collection = f"{container.space}:{container.external_id}.{prop_id}"
                for identifier, value in prop.type.values.items():
                    enum_by_container_property[(container_id, prop_id)].append(
                        DMSInputEnum(
                            collection=collection, value=identifier, name=value.name, description=value.description
                        )
                    )
        return enum_by_container_property



import urllib.parse
from collections.abc import Iterable, Iterator, Set
from functools import cached_property
from typing import cast

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling import DataModelIdentifier
from cognite.client.data_classes.data_modeling.instances import Instance, InstanceSort
from cognite.client.utils.useful_types import SequenceNotStr
from rdflib import RDF, Literal, Namespace, URIRef

from cognite.neat._config import GLOBAL_CONFIG
from cognite.neat._constants import DEFAULT_SPACE_URI, is_readonly_property
from cognite.neat._issues.errors import ResourceRetrievalError
from cognite.neat._shared import Triple
from cognite.neat._utils.collection_ import iterate_progress_bar

from ._base import BaseExtractor
from ._dict import DEFAULT_EMPTY_VALUES, DMSPropertyExtractor


class DMSExtractor(BaseExtractor):
    """Extract data from Cognite Data Fusion DMS instances into Neat.

    Args:
        total_instances_pair_by_view: A dictionary where the key is the view id and the value is a tuple with the total
            number of instances and an iterable of instances.
        limit: The maximum number of items to extract.
        overwrite_namespace: If provided, this will overwrite the space of the extracted items.
        unpack_json: If True, JSON objects will be unpacked into RDF literals.
        empty_values: If unpack_json is True, when unpacking JSON objects, if a key has a value in this set, it will be
            considered as an empty value and skipped.
        str_to_ideal_type: If unpack_json is True, when unpacking JSON objects, if the value is a string, the extractor
            will try to convert it to the ideal type.
    """

    def __init__(
        self,
        total_instances_pair_by_view: dict[dm.ViewId, tuple[int | None, Iterable[Instance]]],
        limit: int | None = None,
        overwrite_namespace: Namespace | None = None,
        unpack_json: bool = False,
        empty_values: Set[str] = DEFAULT_EMPTY_VALUES,
        str_to_ideal_type: bool = False,
    ) -> None:
        self.total_instances_pair_by_view = total_instances_pair_by_view
        self.limit = limit
        self.overwrite_namespace = overwrite_namespace
        self.unpack_json = unpack_json
        self.empty_values = empty_values
        self.str_to_ideal_type = str_to_ideal_type

    @classmethod
    def from_data_model(
        cls,
        client: CogniteClient,
        data_model: DataModelIdentifier,
        limit: int | None = None,
        overwrite_namespace: Namespace | None = None,
        instance_space: str | SequenceNotStr[str] | None = None,
        unpack_json: bool = False,
        str_to_ideal_type: bool = False,
    ) -> "DMSExtractor":
        """Create an extractor from a data model.

        Args:
            client: The Cognite client to use.
            data_model: The data model to extract.
            limit: The maximum number of instances to extract.
            overwrite_namespace: If provided, this will overwrite the space of the extracted items.
            instance_space: The space to extract instances from.
            unpack_json: If True, JSON objects will be unpacked into RDF literals.
        """
        retrieved = client.data_modeling.data_models.retrieve(data_model, inline_views=True)
        if not retrieved:
            raise ResourceRetrievalError(dm.DataModelId.load(data_model), "data model", "Data Model is missing in CDF")
        return cls.from_views(
            client,
            retrieved.latest_version().views,
            limit,
            overwrite_namespace,
            instance_space,
            unpack_json,
            str_to_ideal_type,
        )

    @classmethod
    def from_views(
        cls,
        client: CogniteClient,
        views: Iterable[dm.View],
        limit: int | None = None,
        overwrite_namespace: Namespace | None = None,
        instance_space: str | SequenceNotStr[str] | None = None,
        unpack_json: bool = False,
        str_to_ideal_type: bool = False,
    ) -> "DMSExtractor":
        """Create an extractor from a set of views.

        Args:
            client: The Cognite client to use.
            views: The views to extract.
            limit: The maximum number of instances to extract.
            overwrite_namespace: If provided, this will overwrite the space of the extracted items.
            instance_space: The space to extract instances from.
            unpack_json: If True, JSON objects will be unpacked into RDF literals.
            str_to_ideal_type: If True, when unpacking JSON objects, if the value is a string, the extractor will try to
                convert it to the ideal type.
        """
        total_instances_pair_by_view: dict[dm.ViewId, tuple[int | None, Iterable[Instance]]] = {}
        for view in views:
            instance_iterator = _ViewInstanceIterator(client, view, instance_space)
            total_instances_pair_by_view[view.as_id()] = (instance_iterator.count, instance_iterator)

        return cls(
            total_instances_pair_by_view=total_instances_pair_by_view,
            limit=limit,
            overwrite_namespace=overwrite_namespace,
            unpack_json=unpack_json,
            str_to_ideal_type=str_to_ideal_type,
        )

    def extract(self) -> Iterable[Triple]:
        total_instances = sum(total for total, _ in self.total_instances_pair_by_view.values() if total is not None)
        use_progress_bar = (
            GLOBAL_CONFIG.use_iterate_bar_threshold and total_instances > GLOBAL_CONFIG.use_iterate_bar_threshold
        )

        for view_id, (total, instances) in self.total_instances_pair_by_view.items():
            if total == 0:
                continue
            if use_progress_bar and total is not None:
                instances = iterate_progress_bar(
                    instances,
                    total,
                    f"Extracting instances from {view_id.space}:{view_id.external_id}(version={view_id.version})",
                )

            for count, item in enumerate(instances, 1):
                if self.limit and count > self.limit:
                    break
                yield from self._extract_instance(item)

    def _extract_instance(self, instance: Instance) -> Iterable[Triple]:
        if isinstance(instance, dm.Edge):
            if not instance.properties:
                yield (
                    self._as_uri_ref(instance.start_node),
                    self._as_uri_ref(instance.type),
                    self._as_uri_ref(instance.end_node),
                )
                return
            else:
                # If the edge has properties, we create a node for the edge and connect it to the start and end nodes.
                id_ = self._as_uri_ref(instance)
                yield id_, RDF.type, self._as_uri_ref(instance.type)
                yield (
                    id_,
                    self._as_uri_ref(dm.DirectRelationReference(instance.space, "startNode")),
                    self._as_uri_ref(instance.start_node),
                )
                yield (
                    id_,
                    self._as_uri_ref(dm.DirectRelationReference(instance.space, "endNode")),
                    self._as_uri_ref(instance.end_node),
                )

        elif isinstance(instance, dm.Node):
            id_ = self._as_uri_ref(instance)
            if instance.type:
                type_ = self._as_uri_ref(cast(dm.DirectRelationReference, instance.type))
            elif len(instance.properties) == 1:
                view_id = next(iter(instance.properties.keys()))
                type_ = self._get_namespace(view_id.space)[urllib.parse.quote(view_id.external_id)]
            else:
                type_ = self._get_namespace(instance.space).Node

            yield id_, RDF.type, type_
        else:
            raise NotImplementedError(f"Unknown instance type {type(instance)}")

        if self.overwrite_namespace:
            # If the namespace is overwritten, keep the original space as a property to avoid losing information.
            yield id_, self._get_namespace(instance.space)["space"], Literal(instance.space)

        for view_id, properties in instance.properties.items():
            namespace = self._get_namespace(view_id.space)
            yield from DMSPropertyExtractor(
                id_,
                properties,
                namespace,
                self._as_uri_ref,
                self.empty_values,
                self.str_to_ideal_type,
                self.unpack_json,
            ).extract()

    def _as_uri_ref(self, instance: Instance | dm.DirectRelationReference) -> URIRef:
        return self._get_namespace(instance.space)[urllib.parse.quote(instance.external_id)]

    def _get_namespace(self, space: str) -> Namespace:
        if self.overwrite_namespace:
            return self.overwrite_namespace
        return Namespace(DEFAULT_SPACE_URI.format(space=urllib.parse.quote(space)))


class _ViewInstanceIterator(Iterable[Instance]):
    def __init__(self, client: CogniteClient, view: dm.View, instance_space: str | SequenceNotStr[str] | None = None):
        self.client = client
        self.view = view
        self.instance_space = instance_space

    @cached_property
    def count(self) -> int:
        node_count = edge_count = 0
        if self.view.used_for in ("node", "all"):
            node_count = int(
                self.client.data_modeling.instances.aggregate(
                    view=self.view.as_id(),
                    aggregates=dm.aggregations.Count("externalId"),
                    instance_type="node",
                    space=self.instance_space,
                ).value
            )
        if self.view.used_for in ("edge", "all"):
            edge_count = int(
                self.client.data_modeling.instances.aggregate(
                    view=self.view.as_id(),
                    aggregates=dm.aggregations.Count("externalId"),
                    instance_type="edge",
                    space=self.instance_space,
                ).value
            )
        return node_count + edge_count

    def __iter__(self) -> Iterator[Instance]:
        view_id = self.view.as_id()
        read_only_properties = {
            prop_id
            for prop_id, prop in self.view.properties.items()
            if isinstance(prop, dm.MappedProperty)
            and is_readonly_property(prop.container, prop.container_property_identifier)
        }
        # All nodes and edges with properties
        if self.view.used_for in ("node", "all"):
            # Without a sort, the sort is implicitly by the internal id, as cursoring needs a stable sort.
            # By making the sort be on external_id, Postgres should pick the index
            # that's on (project_id, space, external_id)
            # WHERE deleted_at IS NULL. In other words, avoiding soft deleted instances.
            node_iterable: Iterable[Instance] = self.client.data_modeling.instances(
                chunk_size=None,
                instance_type="node",
                sources=[view_id],
                space=self.instance_space,
                sort=InstanceSort(["node", "externalId"]),
            )
            if read_only_properties:
                node_iterable = self._remove_read_only_properties(node_iterable, read_only_properties, view_id)
            yield from node_iterable

        if self.view.used_for in ("edge", "all"):
            yield from self.client.data_modeling.instances(
                chunk_size=None,
                instance_type="edge",
                sources=[view_id],
                space=self.instance_space,
                sort=InstanceSort(["edge", "externalId"]),
            )

        for prop in self.view.properties.values():
            if isinstance(prop, dm.EdgeConnection):
                if prop.edge_source:
                    # All edges with properties are extracted from the edge source
                    continue
                yield from self.client.data_modeling.instances(
                    chunk_size=None,
                    instance_type="edge",
                    filter=dm.filters.Equals(
                        ["edge", "type"], {"space": prop.type.space, "externalId": prop.type.external_id}
                    ),
                    space=self.instance_space,
                    sort=InstanceSort(["edge", "externalId"]),
                )

    @staticmethod
    def _remove_read_only_properties(
        nodes: Iterable[Instance], read_only_properties: Set[str], view_id: dm.ViewId
    ) -> Iterable[Instance]:
        for node in nodes:
            if properties := node.properties.get(view_id):
                for read_only in read_only_properties:
                    properties.pop(read_only, None)
            yield node



import re
import sys
import typing
from datetime import date, datetime
from typing import Any, ClassVar

from cognite.client.data_classes import data_modeling as dms
from cognite.client.data_classes.data_modeling.data_types import Enum as DMSEnum
from pydantic import BaseModel, Field, model_serializer, model_validator
from pydantic.functional_validators import ModelWrapValidatorHandler
from rdflib import URIRef

from cognite.neat._constants import XML_SCHEMA_NAMESPACE
from cognite.neat._rules._constants import SPLIT_ON_COMMA_PATTERN, SPLIT_ON_EQUAL_PATTERN
from cognite.neat._rules.models.entities._single_value import ClassEntity, UnitEntity

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

# This patterns matches a string that is a data type, with optional content in parentheses.
# For example, it matches "float(unit=power:megaw)" as name="float" and content="unit=power:megaw"
_DATATYPE_PATTERN = re.compile(r"^(?P<name>[^(:]*)(\((?P<content>.+)\))?$")


class DataType(BaseModel):
    python: ClassVar[type]
    dms: ClassVar[type[dms.PropertyType]]
    graphql: ClassVar[str]
    xsd: ClassVar[str]
    sql: ClassVar[str]
    # Repeat all here, just to make mypy happy
    name: typing.Literal[
        "boolean",
        "token",
        "float",
        "double",
        "decimal",
        "integer",
        "nonPositiveInteger",
        "nonNegativeInteger",
        "long",
        "negativeInteger",
        "string",
        "langString",
        "anyURI",
        "normalizedString",
        "dateTime",
        "timestamp",
        "dateTimeStamp",
        "date",
        "plainLiteral",
        "Literal",
        "timeseries",
        "file",
        "sequence",
        "json",
        "enum",
        "",
    ] = ""

    @classmethod
    def load(cls, data: Any) -> Self:
        return cls.model_validate(data)

    def dump(self) -> dict[str, Any]:
        return self.model_dump(by_alias=True)

    @model_validator(mode="wrap")
    def _load(cls, value: Any, handler: ModelWrapValidatorHandler["DataType"]) -> Any:
        if cls is not DataType or isinstance(value, DataType):
            # This is a subclass, let the subclass handle it
            return handler(value)
        elif isinstance(value, str) and (match := _DATATYPE_PATTERN.match(value)):
            name = match.group("name").casefold()
            cls_: type[DataType]
            if name in _DATA_TYPE_BY_DMS_TYPE:
                cls_ = _DATA_TYPE_BY_DMS_TYPE[name]
            elif name in _DATA_TYPE_BY_NAME:
                cls_ = _DATA_TYPE_BY_NAME[name]
            else:
                raise ValueError(f"Unknown data type: {value}") from None
            extra_args: dict[str, Any] = {}
            if content := match.group("content"):
                extra_args = dict(
                    SPLIT_ON_EQUAL_PATTERN.split(pair.strip()) for pair in SPLIT_ON_COMMA_PATTERN.split(content)
                )
                # Todo? Raise warning if extra_args contains keys that are not in the model fields
            instance = cls_(**extra_args)
            created = handler(instance)
            return created
        raise ValueError(f"Cannot load {cls.__name__} from {value}")

    @model_serializer(when_used="unless-none", return_type=str)
    def as_str(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return self._suffix_extra_args(self.model_fields["name"].default)

    def _suffix_extra_args(self, base: str) -> str:
        extra_fields: dict[str, str] = {}
        for field_id, field_ in self.model_fields.items():
            if field_id == "name":
                continue
            value = getattr(self, field_id)
            if value is None:
                continue
            extra_fields[field_.alias or field_id] = str(value)
        if extra_fields:
            content = ",".join(f"{key}={value}" for key, value in sorted(extra_fields.items(), key=lambda x: x[0]))
            return f"{base}({content})"
        return base

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, type(self)) and str(self) == str(other)

    def __hash__(self) -> int:
        return hash(str(self))

    @classmethod
    def is_data_type(cls, value: str) -> bool:
        if match := _DATATYPE_PATTERN.match(value):
            name = match.group("name").casefold()
            return name in _DATA_TYPE_BY_NAME or name in _DATA_TYPE_BY_DMS_TYPE
        return False

    @classmethod
    def as_xml_uri_ref(cls) -> URIRef:
        return XML_SCHEMA_NAMESPACE[cls.xsd]

    @classmethod
    def convert_value(cls, value: Any) -> Any:
        if cls != Boolean:
            return cls.python(value)
        else:
            return value.strip().lower() in {"true", "1", "yes"}


class Boolean(DataType):
    python = bool
    dms = dms.Boolean
    graphql = "Boolean"
    xsd = "boolean"
    sql = "BOOLEAN"
    name: typing.Literal["boolean"] = "boolean"


class Float(DataType):
    python = float
    dms = dms.Float32
    graphql = "Float"
    xsd = "float"
    sql = "FLOAT"

    name: typing.Literal["float"] = "float"
    unit: UnitEntity | None = None


class Double(DataType):
    python = float
    dms = dms.Float64
    graphql = "Float"
    xsd = "double"
    sql = "FLOAT"

    name: typing.Literal["double"] = "double"
    unit: UnitEntity | None = None


class Decimal(DataType):
    python = float
    dms = dms.Float64
    graphql = "Float"
    xsd = "decimal"
    sql = "FLOAT"

    name: typing.Literal["decimal"] = "decimal"
    unit: UnitEntity | None = None


class Integer(DataType):
    python = int
    dms = dms.Int32
    graphql = "Int"
    xsd = "integer"
    sql = "INTEGER"

    name: typing.Literal["integer"] = "integer"


class NonPositiveInteger(DataType):
    python = int
    dms = dms.Int32
    graphql = "Int"
    xsd = "nonPositiveInteger"
    sql = "INTEGER"

    name: typing.Literal["nonPositiveInteger"] = "nonPositiveInteger"


class NonNegativeInteger(DataType):
    python = int
    dms = dms.Int32
    graphql = "Int"
    xsd = "nonNegativeInteger"
    sql = "INTEGER"

    name: typing.Literal["nonNegativeInteger"] = "nonNegativeInteger"


class NegativeInteger(DataType):
    python = int
    dms = dms.Int32
    graphql = "Int"
    xsd = "negativeInteger"
    sql = "INTEGER"

    name: typing.Literal["negativeInteger"] = "negativeInteger"


class Long(DataType):
    python = int
    dms = dms.Int64
    graphql = "Int"
    xsd = "long"
    sql = "BIGINT"

    name: typing.Literal["long"] = "long"


class String(DataType):
    python = str
    dms = dms.Text
    graphql = "String"
    xsd = "string"
    sql = "STRING"

    name: typing.Literal["string"] = "string"


class LangString(DataType):
    python = str
    dms = dms.Text
    graphql = "String"
    xsd = "string"
    sql = "STRING"

    name: typing.Literal["langString"] = "langString"


class AnyURI(DataType):
    python = str
    dms = dms.Text
    graphql = "String"
    xsd = "anyURI"
    sql = "STRING"

    name: typing.Literal["anyURI"] = "anyURI"


class NormalizedString(DataType):
    python = str
    dms = dms.Text
    graphql = "String"
    xsd = "normalizedString"
    sql = "STRING"

    name: typing.Literal["normalizedString"] = "normalizedString"


class Token(DataType):
    python = str
    dms = dms.Text
    graphql = "String"
    xsd = "string"
    sql = "STRING"

    name: typing.Literal["token"] = "token"


class DateTime(DataType):
    python = datetime
    dms = dms.Timestamp
    graphql = "Timestamp"
    xsd = "dateTimeStamp"
    sql = "TIMESTAMP"

    name: typing.Literal["dateTime"] = "dateTime"


class Timestamp(DataType):
    python = datetime
    dms = dms.Timestamp
    graphql = "Timestamp"
    xsd = "dateTimeStamp"
    sql = "TIMESTAMP"

    name: typing.Literal["timestamp"] = "timestamp"


class DateTimeStamp(DataType):
    python = datetime
    dms = dms.Timestamp
    graphql = "Timestamp"
    xsd = "dateTimeStamp"
    sql = "TIMESTAMP"

    name: typing.Literal["dateTimeStamp"] = "dateTimeStamp"


class Date(DataType):
    python = date
    dms = dms.Date
    graphql = "String"
    xsd = "date"
    sql = "DATE"

    name: typing.Literal["date"] = "date"


class PlainLiteral(DataType):
    python = str
    dms = dms.Text
    graphql = "String"
    xsd = "plainLiteral"
    sql = "STRING"

    name: typing.Literal["plainLiteral"] = "plainLiteral"


class Literal(DataType):
    python = str
    dms = dms.Text
    graphql = "String"
    xsd = "string"
    sql = "STRING"

    name: typing.Literal["Literal"] = "Literal"


class Timeseries(DataType):
    python = str
    dms = dms.TimeSeriesReference
    graphql = "TimeSeries"
    xsd = "string"
    sql = "STRING"

    name: typing.Literal["timeseries"] = "timeseries"


class File(DataType):
    python = str
    dms = dms.FileReference
    graphql = "File"
    xsd = "string"
    sql = "STRING"

    name: typing.Literal["file"] = "file"


class Sequence(DataType):
    python = str
    dms = dms.SequenceReference
    graphql = "Sequence"
    xsd = "string"
    sql = "STRING"

    name: typing.Literal["sequence"] = "sequence"


class Json(DataType):
    python = dict
    dms = dms.Json
    graphql = "Json"
    xsd = "json"
    sql = "STRING"

    name: typing.Literal["json"] = "json"


class Enum(DataType):
    python = str
    dms = DMSEnum
    graphql = "Enum"
    xsd = "string"
    sql = "STRING"

    name: typing.Literal["enum"] = "enum"

    collection: ClassEntity
    unknown_value: str | None = Field(None, alias="unknownValue")


_DATA_TYPE_BY_NAME = {cls.model_fields["name"].default.casefold(): cls for cls in DataType.__subclasses__()}
_seen = set()
_DATA_TYPE_BY_DMS_TYPE = {
    cls.dms._type.casefold(): cls
    for cls in DataType.__subclasses__()
    if cls.dms._type not in _seen and not _seen.add(cls.dms._type)  # type: ignore[func-returns-value]
}
del _seen
_XSD_TYPES = {cls_.xsd for cls_ in _DATA_TYPE_BY_NAME.values()}



from pydantic import BaseModel


class _UndefinedType(BaseModel): ...


class _UnknownType(BaseModel):
    def __str__(self) -> str:
        return "#N/A"

    def __hash__(self):
        return hash(str(self))


# This is a trick to make Undefined and Unknown singletons
Undefined = _UndefinedType()
Unknown = _UnknownType()
_PARSE = object()



from cognite.client import CogniteClient
from cognite.client import data_modeling as dm

from cognite.neat._client import NeatClient
from cognite.neat._constants import COGNITE_MODELS
from cognite.neat._graph.transformers import SetType
from cognite.neat._issues import IssueList
from cognite.neat._issues.errors import NeatValueError
from cognite.neat._rules.models import DMSRules
from cognite.neat._rules.transformers import SetIDDMSModel
from cognite.neat._utils.text import humanize_collection

from ._state import SessionState
from .exceptions import NeatSessionError, session_class_wrapper


@session_class_wrapper
class SetAPI:
    """Used to change the name of the data model from a data model id defined by neat to a user specified name."""

    def __init__(self, state: SessionState, verbose: bool) -> None:
        self._state = state
        self._verbose = verbose
        self.instances = SetInstances(state, verbose)

    def data_model_id(self, new_model_id: dm.DataModelId | tuple[str, str, str], name: str | None = None) -> IssueList:
        """Sets the data model ID of the latest verified data model. Set the data model id as a tuple of strings
        following the template (<data_model_space>, <data_model_name>, <data_model_version>).

        Args:
            new_model_id (dm.DataModelId | tuple[str, str, str]): The new data model id.
            name (str, optional): The display name of the data model. If not set, the external ID will be used
                to generate the name.

        Example:
            Set a new data model id:
            ```python
            neat.set.data_model_id(("my_data_model_space", "My_Data_Model", "v1"))
            neat.set.data_model_id(("my_data_model_space", "MyDataModel", "v1"), name="My Data Model")
            ```
        """
        if self._state.rule_store.empty:
            raise NeatSessionError("No rules to set the data model ID.")
        rules = self._state.rule_store.provenance[-1].target_entity.dms
        if isinstance(rules, DMSRules):
            if rules.metadata.as_data_model_id() in COGNITE_MODELS:
                raise NeatSessionError(
                    "Cannot change the data model ID of a Cognite Data Model in NeatSession"
                    " due to temporarily issue with the reverse direct relation interpretation"
                )
        return self._state.rule_transform(SetIDDMSModel(new_model_id))

    def client(self, client: CogniteClient) -> None:
        """Sets the client to be used in the session."""
        self._state.client = NeatClient(client)
        if self._verbose:
            print(f"Client set to {self._state.client.config.project} CDF project.")
        return None


@session_class_wrapper
class SetInstances:
    """Used to change instances"""

    def __init__(self, state: SessionState, verbose: bool) -> None:
        self._state = state
        self._verbose = verbose

    def type_using_property(self, current_type: str, property_type: str, drop_property: bool = True) -> None:
        """Replaces the type of all instances with the value of a property.

        Example:
            All Assets have a property `assetCategory` that we want to use as the type of all asset instances.

            ```python
            neat.set.instances.replace_type("Asset", "assetCategory")
            ```
        """
        type_uri = self._state.instances.store.queries.type_uri(current_type)
        property_uri = self._state.instances.store.queries.property_uri(property_type)

        if not type_uri:
            raise NeatValueError(f"Type {current_type} does not exist in the graph.")
        elif len(type_uri) > 1:
            raise NeatValueError(
                f"{current_type} has multiple ids found in the graph: {humanize_collection(type_uri)}."
            )

        if not property_uri:
            raise NeatValueError(f"Property {property_type} does not exist in the graph.")
        elif len(type_uri) > 1:
            raise NeatValueError(
                f"{property_type} has multiple ids found in the graph: {humanize_collection(property_uri)}."
            )

        if not self._state.instances.store.queries.type_with_property(type_uri[0], property_uri[0]):
            raise NeatValueError(f"Property {property_type} is not defined for type {current_type}.")

        self._state.instances.store.transform(SetType(type_uri[0], property_uri[0], drop_property))

        return None



import colorsys
import random
from typing import Any

import networkx as nx
from IPython.display import HTML, display
from pyvis.network import Network as PyVisNetwork  # type: ignore

from cognite.neat._constants import IN_NOTEBOOK, IN_PYODIDE
from cognite.neat._rules.analysis._base import RulesAnalysis
from cognite.neat._session.exceptions import NeatSessionError
from cognite.neat._utils.io_ import to_directory_compatible
from cognite.neat._utils.rdf_ import remove_namespace_from_uri, uri_display_name

from ._state import SessionState
from .exceptions import session_class_wrapper


@session_class_wrapper
class ShowAPI:
    """Visualise a verified data model or instances contained in the graph store.
    See, for example, `.data_model()` or `.instances()` for more.

    Example:
        Show instances
        ```python
        from cognite.neat import NeatSession
        from cognite.neat import get_cognite_client

        client = get_cognite_client(env_file_path=".env")
        neat = NeatSession(client, storage="oxigraph") # Storage optimised for storage visualisation

        # .... intermediate steps of reading, infering verifying and converting a data model and instances

        neat.show.instances()
        ```

    Example:
        Show data model
        ```python
        neat.show.data_model()
        ```
    """

    def __init__(self, state: SessionState) -> None:
        self._state = state
        self.data_model = ShowDataModelAPI(self._state)
        self.instances = ShowInstanceAPI(self._state)


@session_class_wrapper
class ShowBaseAPI:
    def __init__(self, state: SessionState) -> None:
        self._state = state

    def _generate_visualization(self, di_graph: nx.DiGraph, name: str) -> Any:
        if not IN_NOTEBOOK:
            raise NeatSessionError("Visualization is only available in Jupyter notebooks!")

        net = PyVisNetwork(
            notebook=IN_NOTEBOOK,
            cdn_resources="remote",
            directed=True,
            height="750px",
            width="100%",
            select_menu=IN_NOTEBOOK,
        )

        # Change the plotting layout
        net.repulsion(
            node_distance=100,
            central_gravity=0.3,
            spring_length=200,
            spring_strength=0.05,
            damping=0.09,
        )

        net.from_nx(di_graph)
        if IN_PYODIDE:
            net.write_html(name)
            return display(HTML(name))

        else:
            return net.show(name)


@session_class_wrapper
class ShowDataModelAPI(ShowBaseAPI):
    """Visualises the verified data model."""

    def __init__(self, state: SessionState) -> None:
        super().__init__(state)
        self._state = state
        self.provenance = ShowDataModelProvenanceAPI(self._state)
        self.implements = ShowDataModelImplementsAPI(self._state)

    def __call__(self) -> Any:
        if self._state.rule_store.empty:
            raise NeatSessionError("No data model available. Try using [bold].read[/bold] to read a data model.")

        last_target = self._state.rule_store.provenance[-1].target_entity
        rules = last_target.dms or last_target.information
        analysis = RulesAnalysis(dms=last_target.dms, information=last_target.information)

        if last_target.dms is not None:
            di_graph = analysis._dms_di_graph(format="data-model")
        else:
            di_graph = analysis._info_di_graph(format="data-model")

        identifier = to_directory_compatible(str(rules.metadata.identifier))
        name = f"{identifier}.html"
        return self._generate_visualization(di_graph, name)


@session_class_wrapper
class ShowDataModelImplementsAPI(ShowBaseAPI):
    def __init__(self, state: SessionState) -> None:
        super().__init__(state)
        self._state = state

    def __call__(self) -> Any:
        if self._state.rule_store.empty:
            raise NeatSessionError("No data model available. Try using [bold].read[/bold] to read a data model.")

        last_target = self._state.rule_store.provenance[-1].target_entity
        rules = last_target.dms or last_target.information
        analysis = RulesAnalysis(dms=last_target.dms, information=last_target.information)

        if last_target.dms is not None:
            di_graph = analysis._dms_di_graph(format="implements")
        else:
            di_graph = analysis._info_di_graph(format="implements")
        identifier = to_directory_compatible(str(rules.metadata.identifier))
        name = f"{identifier}_implements.html"
        return self._generate_visualization(di_graph, name)


@session_class_wrapper
class ShowDataModelProvenanceAPI(ShowBaseAPI):
    """Visualises the provenance or steps that have been executed in the NeatSession."""

    def __init__(self, state: SessionState) -> None:
        super().__init__(state)
        self._state = state

    def __call__(self) -> Any:
        if not self._state.rule_store.provenance:
            raise NeatSessionError("No data model available. Try using [bold].read[/bold] to load data model.")

        di_graph = self._generate_dm_provenance_di_graph_and_types()
        unique_hash = self._state.rule_store.calculate_provenance_hash(shorten=True)
        return self._generate_visualization(di_graph, name=f"data_model_provenance_{unique_hash}.html")

    def _generate_dm_provenance_di_graph_and_types(self) -> nx.DiGraph:
        di_graph = nx.DiGraph()
        hex_colored_types = _generate_hex_color_per_type(["Agent", "Entity", "Activity", "Export", "Pruned"])

        for change in self._state.rule_store.provenance:
            source = uri_display_name(change.source_entity.id_)
            target = uri_display_name(change.target_entity.id_)
            agent = uri_display_name(change.agent.id_)

            di_graph.add_node(
                source,
                label=source,
                type="Entity",
                title="Entity",
                color=hex_colored_types["Entity"],
            )

            di_graph.add_node(
                target,
                label=target,
                type="Entity",
                title="Entity",
                color=hex_colored_types["Entity"],
            )

            di_graph.add_node(
                agent,
                label=agent,
                type="Agent",
                title="Agent",
                color=hex_colored_types["Agent"],
            )

            di_graph.add_edge(source, agent, label="used", color="grey")
            di_graph.add_edge(agent, target, label="generated", color="grey")

        for source_id, exports in self._state.rule_store.exports_by_source_entity_id.items():
            source_shorten = uri_display_name(source_id)
            for export in exports:
                export_id = uri_display_name(export.target_entity.id_)
                di_graph.add_node(
                    export_id,
                    label=export_id,
                    type="Export",
                    title="Export",
                    color=hex_colored_types["Export"],
                )
                di_graph.add_edge(source_shorten, export_id, label="exported", color="grey")

        return di_graph


@session_class_wrapper
class ShowInstanceAPI(ShowBaseAPI):
    """Visualise the instances contained in the graph store."""

    def __init__(self, state: SessionState) -> None:
        super().__init__(state)
        self._state = state

    def __call__(self) -> Any:
        if IN_PYODIDE:
            raise NeatSessionError(
                "Instances visualization not available if neat is run in PYODIDE as "
                "PYODIDE dues not support oxigraph storage for NeatSession."
                "Try running neat in regular Jupyter notebook and set [bold]NeatSession(storage='oxigraph')[/bold]."
            )

        if not self._state.instances.store_type == "oxigraph":
            raise NeatSessionError(
                "Visualization is only available for Oxigraph store. "
                'Try setting [bold]NeatSession(storage="oxigraph")[/bold] enable Oxigraph store.'
            )

        if not self._state.instances.store.dataset:
            raise NeatSessionError("No instances available. Try using [bold].read[/bold] to load instances.")

        di_graph = self._generate_instance_di_graph_and_types()
        return self._generate_visualization(di_graph, name="instances.html")

    def _generate_instance_di_graph_and_types(self) -> nx.DiGraph:
        query = """
        SELECT ?s ?p ?o ?ts ?to WHERE {
            ?s ?p ?o .
            FILTER(isIRI(?o))  # Example filter to check if ?o is an IRI (object type)
            FILTER(BOUND(?o))
            FILTER(?p != rdf:type)

            ?s a ?ts .
            ?o a ?to .
        }
        LIMIT 200
        """

        di_graph = nx.DiGraph()

        types = [type_ for type_, _ in self._state.instances.store.queries.summarize_instances()]
        hex_colored_types = _generate_hex_color_per_type(types)

        for (  # type: ignore
            subject,
            property_,
            object,
            subject_type,
            object_type,
        ) in self._state.instances.store.dataset.query(query):
            subject = remove_namespace_from_uri(subject)
            property_ = remove_namespace_from_uri(property_)
            object = remove_namespace_from_uri(object)
            subject_type = remove_namespace_from_uri(subject_type)
            object_type = remove_namespace_from_uri(object_type)

            di_graph.add_node(
                subject,
                label=subject,
                type=subject_type,
                title=subject_type,
                color=hex_colored_types[subject_type],
            )
            di_graph.add_node(
                object,
                label=object,
                type=object_type,
                title=object_type,
                color=hex_colored_types[object_type],
            )
            di_graph.add_edge(subject, object, label=property_, color="grey")

        return di_graph


def _generate_hex_color_per_type(types: list[str]) -> dict:
    hex_colored_types = {}
    random.seed(381)
    for type_ in types:
        hue = random.random()
        saturation = random.uniform(0.5, 1.0)
        lightness = random.uniform(0.4, 0.6)
        rgb = colorsys.hls_to_rgb(hue, lightness, saturation)
        hex_color = f"#{int(rgb[0] * 255):02x}{int(rgb[1] * 255):02x}{int(rgb[2] * 255):02x}"
        hex_colored_types[type_] = hex_color
    return hex_colored_types



from cognite.neat._issues import IssueList
from cognite.neat._rules.models.mapping import load_classic_to_core_mapping
from cognite.neat._rules.transformers import (
    AsParentPropertyId,
    ChangeViewPrefix,
    IncludeReferenced,
    RuleMapper,
    VerifiedRulesTransformer,
)

from ._state import SessionState
from .exceptions import NeatSessionError, session_class_wrapper


@session_class_wrapper
class MappingAPI:
    def __init__(self, state: SessionState):
        self.data_model = DataModelMappingAPI(state)


@session_class_wrapper
class DataModelMappingAPI:
    def __init__(self, state: SessionState):
        self._state = state

    def classic_to_core(self, company_prefix: str | None = None, use_parent_property_name: bool = True) -> IssueList:
        """Map classic types to core types.

        Note this automatically creates an extended CogniteCore model.

        Args:
            company_prefix: Prefix used for all extended CogniteCore types.
            use_parent_property_name: Whether to use the parent property name in the extended CogniteCore model.
                See below for more information.

        If you extend CogniteAsset, with for example, ClassicAsset. You will map the property `parentId` to `parent`.
        If you set `user_parent_property_name` to True, the `parentId` will be renamed to `parent` after the
        mapping is done. If you set it to False, the property will remain `parentId`.

        Example:
            ```python
            neat.mapping.classic_to_core(company_prefix="WindFarmX", use_parent_property_name=True)
            ```
        """
        if self._state.rule_store.empty:
            raise NeatSessionError("No rules to map")
        last_entity = self._state.rule_store.provenance[-1].target_entity
        if last_entity.dms is None:
            raise NeatSessionError("Data model not converted to DMS. Try running `neat.convert('dms')` first.")
        rules = last_entity.dms
        if self._state.client is None:
            raise NeatSessionError("Client is required to map classic to core")

        transformers: list[VerifiedRulesTransformer] = []
        if company_prefix:
            transformers.append(ChangeViewPrefix("Classic", company_prefix))
        transformers.extend(
            [
                RuleMapper(load_classic_to_core_mapping(company_prefix, rules.metadata.space, rules.metadata.version)),
                IncludeReferenced(self._state.client),
            ]
        )
        if use_parent_property_name:
            transformers.append(AsParentPropertyId(self._state.client))
        return self._state.rule_transform(*transformers)



from cognite.client import ClientConfig, CogniteClient

from cognite.neat._utils.auth import _CLIENT_NAME

from ._api.data_modeling_loaders import DataModelLoaderAPI
from ._api.schema import SchemaAPI


class NeatClient(CogniteClient):
    def __init__(self, config: ClientConfig | CogniteClient | None = None) -> None:
        if isinstance(config, CogniteClient):
            config = config.config
        super().__init__(config=config)
        if self._config is not None:
            self._config.client_name = _CLIENT_NAME
        self.loaders = DataModelLoaderAPI(self)
        self.schema = SchemaAPI(self)



import re
from collections.abc import Iterable
from typing import Any, Literal, TypeAlias, overload

from cognite.client.utils.useful_types import SequenceNotStr
from pydantic import HttpUrl, TypeAdapter, ValidationError
from rdflib import Graph, Namespace, URIRef
from rdflib import Literal as RdfLiteral

from cognite.neat._constants import SPACE_URI_PATTERN

Triple: TypeAlias = tuple[URIRef, URIRef, RdfLiteral | URIRef]


@overload
def remove_namespace_from_uri(
    URI: URIRef | str,
    *,
    special_separator: str = "#_",
    validation: Literal["full", "prefix"] = "prefix",
) -> str: ...


@overload
def remove_namespace_from_uri(
    URI: SequenceNotStr[URIRef | str],
    *,
    special_separator: str = "#_",
    validation: Literal["full", "prefix"] = "prefix",
) -> list[str]: ...


def remove_namespace_from_uri(
    URI: URIRef | str | SequenceNotStr[URIRef | str],
    *,
    special_separator: str = "#_",
    validation: Literal["full", "prefix"] = "prefix",
) -> str | list[str]:
    """Removes namespace from URI

    Args
        URI: URIRef | str
            URI of an entity
        special_separator : str
            Special separator to use instead of # or / if present in URI
            Set by default to "#_" which covers special client use case
        validation: str
            Validation type to use for URI. If set to "full", URI will be validated using pydantic
            If set to "prefix", only check if URI starts with http or https will be made

    Returns
        Entities id without namespace

    Examples:

        >>> remove_namespace_from_uri("http://www.example.org/index.html#section2")
        'section2'
        >>> remove_namespace_from_uri("http://www.example.org/index.html#section2", "http://www.example.org/index.html#section3")
        ('section2', 'section3')
    """
    is_single = False
    uris: Iterable[str | URIRef]
    if isinstance(URI, str | URIRef):
        uris = (URI,)
        is_single = True
    elif isinstance(URI, SequenceNotStr):
        # Assume that all elements in the tuple are of the same type following type hint
        uris = URI
    else:
        raise TypeError(f"URI must be of type URIRef or str, got {type(URI)}")

    output = []
    for u in uris:
        if validation == "full":
            try:
                _ = TypeAdapter(HttpUrl).validate_python(u)
                output.append(u.split(special_separator if special_separator in u else "#" if "#" in u else "/")[-1])
            except ValidationError:
                output.append(str(u))
        else:
            if u.lower().startswith("http"):
                output.append(u.split(special_separator if special_separator in u else "#" if "#" in u else "/")[-1])
            else:
                output.append(str(u))

    return output[0] if is_single else output


def get_namespace(URI: URIRef, special_separator: str = "#_") -> str:
    """Removes namespace from URI

    Parameters
    ----------
    URI : URIRef
        URI of an entity
    special_separator : str
        Special separator to use instead of # or / if present in URI
        Set by default to "#_" which covers special client use case

    Returns
    -------
    str
        Entity id without namespace
    """
    return split_uri(URI, special_separator)[0]


def namespace_as_space(namespace: str) -> str | None:
    if match := SPACE_URI_PATTERN.match(namespace):
        return match.group("space")
    return None


def split_uri(URI: URIRef, special_separator: str = "#_") -> tuple[str, str]:
    """Splits URI into namespace and entity name

    Parameters
    ----------
    URI : URIRef
        URI of an entity
    special_separator : str
        Special separator to use instead of # or / if present in URI
        Set by default to "#_" which covers special client use case

    Returns
    -------
    tuple[str, str]
        Tuple of namespace and entity name
    """
    if special_separator in URI:
        namespace, rest = URI.split(special_separator, maxsplit=1)
        namespace += special_separator
    elif "#" in URI:
        namespace, rest = URI.split("#", maxsplit=1)
        namespace += "#"
    else:
        namespace, rest = URI.rsplit("/", maxsplit=1)
        namespace += "/"
    return namespace, rest


def as_neat_compliant_uri(uri: URIRef) -> URIRef:
    namespace = get_namespace(uri)
    id_ = remove_namespace_from_uri(uri)
    compliant_uri = re.sub(r"[^a-zA-Z0-9-_.]", "", id_)
    return URIRef(f"{namespace}{compliant_uri}")


def convert_rdflib_content(content: RdfLiteral | URIRef | dict | list, remove_namespace: bool = False) -> Any:
    if isinstance(content, RdfLiteral):
        return content.toPython()
    elif isinstance(content, URIRef):
        return remove_namespace_from_uri(content) if remove_namespace else content.toPython()
    elif isinstance(content, dict):
        return {key: convert_rdflib_content(value, remove_namespace) for key, value in content.items()}
    elif isinstance(content, list):
        return [convert_rdflib_content(item, remove_namespace) for item in content]
    else:
        return content


def uri_to_short_form(URI: URIRef, prefixes: dict[str, Namespace]) -> str | URIRef:
    """Returns the short form of a URI if its namespace is present in the prefixes dict,
    otherwise returns the URI itself

    Args:
        URI: URI to be converted to form prefix:entityName
        prefixes: dict of prefixes

    Returns:
        shortest form of the URI if its namespace is present in the prefixes dict,
        otherwise returns the URI itself
    """
    uris: set[str | URIRef] = {URI}
    for prefix, namespace in prefixes.items():
        if URI.startswith(namespace):
            uris.add(f"{prefix}:{URI.replace(namespace, '')}")
    return min(uris, key=len)


def _traverse(hierarchy: dict, graph: dict, names: list[str]) -> dict:
    """traverse the graph and return the hierarchy"""
    for name in names:
        hierarchy[name] = _traverse({}, graph, graph[name])
    return hierarchy


def get_inheritance_path(child: Any, child_parent: dict[Any, set[Any]]) -> list[Any]:
    """Returns the inheritance path for a given child

    Args:
        child: Child class
        child_parent: Dictionary of child to parent relationship

    Returns:
        Inheritance path for a given child

    !!! note "No Circular Inheritance"
        This method assumes that the child_parent dictionary is a tree and does not contain any cycles.
    """
    path: list[Any] = []
    if child in child_parent:
        path.extend(child_parent[child])
        for parent in child_parent[child]:
            path.extend(get_inheritance_path(parent, child_parent))
    return path


def add_triples_in_batch(graph: Graph, triples: Iterable[Triple], batch_size: int = 10_000) -> None:
    """Adds triples to the graph store in batches.

    Args:
        triples: list of triples to be added to the graph store
        batch_size: Batch size of triples per commit, by default 10_000
        verbose: Verbose mode, by default False
    """

    commit_counter = 0
    number_of_written_triples = 0

    def check_commit(force_commit: bool = False):
        """Commit nodes to the graph if batch counter is reached or if force_commit is True"""
        nonlocal commit_counter
        nonlocal number_of_written_triples
        if force_commit:
            number_of_written_triples += commit_counter
            graph.commit()
            return
        commit_counter += 1
        if commit_counter >= batch_size:
            number_of_written_triples += commit_counter
            graph.commit()
            commit_counter = 0

    for triple in triples:
        graph.add(triple)
        check_commit()

    check_commit(force_commit=True)


def remove_triples_in_batch(graph: Graph, triples: Iterable[Triple], batch_size: int = 10_000) -> None:
    """Removes triples from the graph store in batches.

    Args:
        triples: list of triples to be removed from the graph store
        batch_size: Batch size of triples per commit, by default 10_000
    """
    batch_count = 0

    def check_commit(force_commit: bool = False):
        """Commit nodes to the graph if batch counter is reached or if force_commit is True"""
        nonlocal batch_count
        batch_count += 1
        if force_commit or batch_count >= batch_size:
            graph.commit()
            batch_count = 0
            return

    for triple in triples:
        graph.remove(triple)
        check_commit()
    check_commit(force_commit=True)


def remove_instance_ids_in_batch(graph: Graph, instance_ids: Iterable[URIRef], batch_size: int = 1_000) -> None:
    """Removes all triples related to the given instances in the graph store in batches.

    Args:
        graph: The graph store to remove triples from
        instance_ids:  list of instances to remove triples from
        batch_size:  Batch size of triples per commit, by default 10_000

    """
    batch_count = 0

    def check_commit(force_commit: bool = False):
        """Commit nodes to the graph if batch counter is reached or if force_commit is True"""
        nonlocal batch_count
        batch_count += 1
        if force_commit or batch_count >= batch_size:
            graph.commit()
            batch_count = 0
            return

    for instance_id in instance_ids:
        graph.remove((instance_id, None, None))
        check_commit()

    check_commit(force_commit=True)


def uri_display_name(thing: URIRef) -> str:
    if "https://cognitedata.com/dms/data-model/" in thing:
        return "DMS(" + ",".join(thing.replace("https://cognitedata.com/dms/data-model/", "").split("/")) + ")"
    elif "http://purl.org/cognite/neat/data-model/" in thing:
        return "NEAT(" + ",".join(thing.replace("http://purl.org/cognite/neat/data-model/", "").split("/")) + ")"
    return remove_namespace_from_uri(thing)



import difflib
from collections.abc import Callable, Set
from typing import Literal, overload

import pandas as pd

from cognite.neat._constants import IN_NOTEBOOK
from cognite.neat._issues import IssueList
from cognite.neat._utils.upload import UploadResult, UploadResultCore, UploadResultList

from ._state import SessionState
from .exceptions import session_class_wrapper

try:
    from rich.markdown import Markdown as RichMarkdown

    RICH_AVAILABLE = True
except ImportError:
    RICH_AVAILABLE = False


@session_class_wrapper
class InspectAPI:
    """Inspect issues or outcomes after performing operations with NeatSession.
    To inspect properties of the current data model, try out `.properties()`.

    Example:
        Inspect issues
        ```python
        neat.inspect.issues()
        ```

    Example:
        Inspect outcome after writing a data model
        ```python
        neat.inspect.outcome.data_model()
        ```

    Example:
        Inspect outcome after writing instances
        ```python
        neat.inspect.outcome.instances()
        ```
    """

    def __init__(self, state: SessionState) -> None:
        self._state = state
        self.issues = InspectIssues(state)
        self.outcome = InspectOutcome(state)

    def properties(self) -> pd.DataFrame:
        """Returns the properties of the current data model.

        Example:
            Inspect properties of the current data model
            ```python
            # From an active NeatSession
            neat.inspect.properties
            ```
        """
        if self._state.rule_store.empty:
            return pd.DataFrame()
        last_entity = self._state.rule_store.provenance[-1].target_entity
        if last_entity.dms:
            df = last_entity.dms.properties.to_pandas()
        else:
            df = last_entity.information.properties.to_pandas()
        df.drop(columns=["neatId"], errors="ignore", inplace=True)
        return df

    def views(self) -> pd.DataFrame:
        if self._state.rule_store.empty:
            return pd.DataFrame()
        last_entity = self._state.rule_store.provenance[-1].target_entity
        if last_entity.dms is None:
            return pd.DataFrame()
        df = last_entity.dms.views.to_pandas()
        df.drop(columns=["neatId"], errors="ignore", inplace=True)
        return df


@session_class_wrapper
class InspectIssues:
    """Inspect issues of the current data model."""

    def __init__(self, state: SessionState) -> None:
        self._state = state
        self._max_display = 50

    @overload
    def __call__(
        self,
        search: str | None = None,
        include: Literal["all", "errors", "warning"] | Set[Literal["all", "errors", "warning"]] = "all",
        return_dataframe: Literal[True] = (False if IN_NOTEBOOK else True),  # type: ignore[assignment]
    ) -> pd.DataFrame: ...

    @overload
    def __call__(
        self,
        search: str | None = None,
        include: Literal["all", "errors", "warning"] | Set[Literal["all", "errors", "warning"]] = "all",
        return_dataframe: Literal[False] = (False if IN_NOTEBOOK else True),  # type: ignore[assignment]
    ) -> None: ...

    def __call__(
        self,
        search: str | None = None,
        include: Literal["all", "errors", "warning"] | Set[Literal["all", "errors", "warning"]] = "all",
        return_dataframe: bool = (False if IN_NOTEBOOK else True),  # type: ignore[assignment]
    ) -> pd.DataFrame | None:
        """Returns the issues of the current data model."""
        issues = self._state.rule_store.last_issues
        if issues is None and self._state.instances.store.provenance:
            last_change = self._state.instances.store.provenance[-1]
            issues = last_change.target_entity.issues
        elif issues is None:
            self._print("No issues found.")
            return pd.DataFrame() if return_dataframe else None
        include_set = {include} if isinstance(include, str) else include
        if "all" in include_set:
            include_set = {"errors", "warning"}
        if "warning" not in include_set:
            issues = issues.errors
        if "errors" not in include_set:
            issues = issues.warnings

        if issues and search is not None:
            unique_types = {type(issue).__name__ for issue in issues}
            closest_match = set(difflib.get_close_matches(search, unique_types))
            issues = IssueList([issue for issue in issues if type(issue).__name__ in closest_match])

        issue_str = "\n".join(
            [
                f"  * **{type(issue).__name__}**: {issue.as_message(include_type=False)}"
                for issue in issues[: self._max_display]
            ]
            + ([] if len(issues) <= 50 else [f"  * ... {len(issues) - self._max_display} more"])
        )
        markdown_str = f"### {len(issues)} issues found\n\n{issue_str}"
        if IN_NOTEBOOK:
            from IPython.display import Markdown, display

            display(Markdown(markdown_str))
        elif RICH_AVAILABLE:
            from rich import print as rprint

            rprint(RichMarkdown(markdown_str))

        if return_dataframe:
            return issues.to_pandas()
        return None

    def _print(self, message: str) -> None:
        if IN_NOTEBOOK:
            from IPython.display import Markdown, display

            display(Markdown(message))
        else:
            print(message)

    def __repr__(self) -> str:
        return self.__repr_html__()

    def __repr_html__(self) -> str:
        return (
            "Inspect issues by calling .inspect.issues() or "
            "search for specific issues by calling .inspect.issues('MyTypeWarning')."
        )


@session_class_wrapper
class InspectOutcome:
    """
    Inspect the outcome after writing a Data Model and Instances to CDF.
    """

    def __init__(self, state: SessionState) -> None:
        self.data_model = InspectUploadOutcome(lambda: state.rule_store.last_outcome)
        self.instances = InspectUploadOutcome(lambda: state.instances.last_outcome)


@session_class_wrapper
class InspectUploadOutcome:
    def __init__(self, get_last_outcome: Callable[[], UploadResultList]) -> None:
        self._get_last_outcome = get_last_outcome
        self._max_display = 50

    @staticmethod
    def _as_set(value: str | list[str] | None) -> set[str] | None:
        if value is None:
            return None
        if isinstance(value, str):
            return {value}
        return set(value)

    @overload
    def __call__(
        self,
        name: str | list[str] | None = None,
        has_errors: bool = False,
        has_issues: bool = False,
        return_dataframe: Literal[False] = (False if IN_NOTEBOOK else True),  # type: ignore[assignment]
    ) -> None: ...

    @overload
    def __call__(
        self,
        name: str | list[str] | None = None,
        has_errors: bool = False,
        has_issues: bool = False,
        return_dataframe: Literal[True] = (False if IN_NOTEBOOK else True),  # type: ignore[assignment]
    ) -> pd.DataFrame: ...

    def __call__(
        self,
        name: str | list[str] | None = None,
        has_errors: bool = False,
        has_issues: bool = False,
        return_dataframe: bool = (False if IN_NOTEBOOK else True),  # type: ignore[assignment]
    ) -> pd.DataFrame | None:
        """Returns the outcome of the last upload."""
        outcome = self._get_last_outcome()
        name_set = self._as_set(name)

        def outcome_filter(item: UploadResultCore) -> bool:
            nonlocal name_set
            if name_set and item.name not in name_set:
                return False
            if has_errors and not item.error_messages:
                return False
            if has_issues and not item.issues:
                return False
            return True

        outcome = UploadResultList([item for item in outcome if outcome_filter(item)])
        if IN_NOTEBOOK:
            from IPython.display import Markdown, display

            lines: list[str] = []
            for line_no, item in enumerate(outcome):
                lines.append(f"### {item.name}")
                if unique_errors := set(item.error_messages):
                    lines.append("#### Errors")
                    for error in unique_errors:
                        lines.append(f"  * {error}")
                if unique_issue_messages := set([issue.as_message() for issue in item.issues]):
                    lines.append("#### Issues")
                    for issue in unique_issue_messages:
                        lines.append(f"  * {issue}")
                if isinstance(item, UploadResult):
                    dumped = item.dump(aggregate=False)
                    for key, value in dumped.items():
                        if key in ["name", "error_messages", "issues"]:
                            continue
                        lines.append(f"#### {key}")
                        if isinstance(value, list):
                            total = len(value)
                            for i, v in enumerate(value):
                                if key in ["created", "updated", "changed"]:
                                    if i < 50:
                                        lines.append(f"  * {v}")
                                    elif i == 50 and total > 50:
                                        lines.append(f"  * ... {total - 50} more")
                                    elif i == 50 and total == 50:
                                        lines.append(f"  * {v}")
                                else:
                                    lines.append(f"  * {v}")

                        else:
                            lines.append(f"  * {value}")

                if line_no >= self._max_display:
                    lines.append(f"### ... {len(outcome) - self._max_display} more")
                    break

            display(Markdown("\n".join(lines)))

        if return_dataframe:
            return outcome.to_pandas()
        return None



from collections.abc import Iterable
from pathlib import Path

from cognite.client import CogniteClient
from cognite.client.data_classes import DataSet, DataSetList
from cognite.client.utils.useful_types import SequenceNotStr

from ._base import ClassicCDFBaseExtractor, InstanceIdPrefix, T_CogniteResource


class DataSetExtractor(ClassicCDFBaseExtractor[DataSet]):
    """Extract DataSets from Cognite Data Fusions into Neat."""

    _default_rdf_type = "DataSet"
    _instance_id_prefix = InstanceIdPrefix.data_set

    @classmethod
    def _from_dataset(
        cls,
        client: CogniteClient,
        data_set_external_id: SequenceNotStr[str],  # type: ignore[override]
    ) -> tuple[int | None, Iterable[DataSet]]:
        items = client.data_sets.retrieve_multiple(external_ids=data_set_external_id)
        return len(items), items

    @classmethod
    def _from_hierarchy(
        cls, client: CogniteClient, root_asset_external_id: str
    ) -> tuple[int | None, Iterable[T_CogniteResource]]:
        raise NotImplementedError("DataSets do not have a hierarchy.")

    @classmethod
    def _from_file(cls, file_path: str | Path) -> tuple[int | None, Iterable[DataSet]]:
        data_sets = DataSetList.load(Path(file_path).read_text())
        return len(data_sets), data_sets



from dataclasses import dataclass
from typing import Generic

from cognite.neat._issues._base import ResourceType

from ._resources import ResourceError, T_Identifier, T_ReferenceIdentifier


@dataclass(unsafe_hash=True)
class PropertyError(ResourceError[T_Identifier]):
    """Base class for property errors {resource_type} with identifier {identifier}.{property_name}"""

    property_name: str


@dataclass(unsafe_hash=True)
class PropertyNotFoundError(PropertyError, Generic[T_Identifier, T_ReferenceIdentifier]):
    """The {resource_type} with identifier {identifier} does not have a property {property_name}"""

    extra = "referred to by {referred_type} {referred_by} does not exist"
    fix = "Ensure the {resource_type} {identifier} has a {property_name} property"

    referred_by: T_ReferenceIdentifier | None = None
    referred_type: ResourceType | None = None


@dataclass(unsafe_hash=True)
class PropertyTypeNotSupportedError(PropertyError[T_Identifier]):
    """The {resource_type} with identifier {identifier} has a property {property_name}
    of unsupported type {property_type}"""

    property_type: str


@dataclass(unsafe_hash=True)
class ReversedConnectionNotFeasibleError(PropertyError[T_Identifier]):
    """The {resource_type} {identifier}.{property_name} cannot be created: {reason}"""

    reason: str


# This is a generic error that should be used sparingly
@dataclass(unsafe_hash=True)
class PropertyDefinitionError(PropertyError[T_Identifier]):
    """Invalid property definition for {resource_type} {identifier}.{property_name}: {reason}"""

    reason: str


@dataclass(unsafe_hash=True)
class PropertyDefinitionDuplicatedError(PropertyError[T_Identifier]):
    """The {resource_type} with identifier {identifier} has multiple definitions for the property {property_name}
    with values {property_values}
    """

    extra = "in locations {locations} with name {location_name}"

    property_values: frozenset[str | int | float | bool | None | tuple[str | int | float | bool | None, ...]]
    locations: tuple[str | int, ...] | None = None
    location_name: str | None = None


@dataclass(unsafe_hash=True)
class PropertyMappingDuplicatedError(PropertyError[T_Identifier], Generic[T_Identifier, T_ReferenceIdentifier]):
    """The {resource_type} with identifier {identifier}.{property_name} is mapped to by: {mappings}. Ensure
    that only one {mapping_type} maps to {resource_type} {identifier}.{property_name}"""

    mappings: frozenset[T_ReferenceIdentifier]
    mapping_type: ResourceType



import re
from collections.abc import Mapping
from pathlib import Path
from typing import TYPE_CHECKING

from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.ids import DataModelId
from rdflib import DC, DCTERMS, FOAF, OWL, RDF, RDFS, SH, SKOS, XSD, Namespace, URIRef
from rdflib.namespace import DefinedNamespace

from cognite import neat

if TYPE_CHECKING:
    from cognite.neat._rules.models.dms import DMSProperty


def _is_in_notebook() -> bool:
    try:
        from IPython import get_ipython

        if "IPKernelApp" not in get_ipython().config:  # pragma: no cover
            return False
    except ImportError:
        return False
    except AttributeError:
        return False
    return True


def _is_in_browser() -> bool:
    try:
        from pyodide.ffi import IN_BROWSER  # type: ignore [import-not-found]
    except ModuleNotFoundError:
        return False
    return IN_BROWSER


IN_PYODIDE = _is_in_browser()
IN_NOTEBOOK = _is_in_notebook() or IN_PYODIDE


PACKAGE_DIRECTORY = Path(neat.__file__).parent
COGNITE_MODELS = (
    DataModelId("cdf_cdm", "CogniteCore", "v1"),
    DataModelId("cdf_idm", "CogniteProcessIndustries", "v1"),
)
COGNITE_SPACES = frozenset(
    {model.space for model in COGNITE_MODELS}
    | {
        "cdf_360_image_schema",
        "cdf_3d_schema",
        "cdf_apm",
        "cdf_apps_shared",
        "cdf_cdm",
        "cdf_cdm_3d",
        "cdf_cdm_units",
        "cdf_classic",
        "cdf_core",
        "cdf_extraction_extensions",
        "cdf_idm",
        "cdf_industrial_canvas",
        "cdf_infield",
        "cdf_time_series_data",
        "cdf_units",
    }
)
DMS_LISTABLE_PROPERTY_LIMIT = 1000

EXAMPLE_RULES = PACKAGE_DIRECTORY / "_rules" / "examples"
EXAMPLE_GRAPHS = PACKAGE_DIRECTORY / "_graph" / "examples"
EXAMPLE_WORKFLOWS = PACKAGE_DIRECTORY / "_workflows" / "examples"

DEFAULT_SPACE_URI = "http://purl.org/cognite/space/{space}#"
SPACE_URI_PATTERN = re.compile(r"http://purl.org/cognite/space/(?P<space>[^#]+)#$")
DEFAULT_RAW_URI = "http://purl.org/cognite/raw#"
DEFAULT_NAMESPACE = Namespace("http://purl.org/cognite/neat/")
CDF_NAMESPACE = Namespace("https://cognitedata.com/")
DEFAULT_BASE_URI = URIRef(DEFAULT_NAMESPACE)
CLASSIC_CDF_NAMESPACE = Namespace("http://purl.org/cognite/cdf-classic#")
XML_SCHEMA_NAMESPACE = Namespace("http://www.w3.org/2001/XMLSchema#")


class NEAT(DefinedNamespace):
    """
    NEAT internal data model used for internal purposes of the NEAT library

    """

    _fail = True
    _NS = Namespace("http://thisisneat.io/internal/")

    UnknownType: URIRef  # Unknown type used to express that the type of a subject is unknown
    EmptyType: URIRef  # Empty type used to express that the type of a subject is empty


def get_default_prefixes_and_namespaces() -> dict[str, Namespace]:
    return {
        "owl": OWL._NS,
        "rdf": RDF._NS,
        "rdfs": RDFS._NS,
        "dcterms": DCTERMS._NS,
        "dc": DC._NS,
        "skos": SKOS._NS,
        "sh": SH._NS,
        "xsd": XSD._NS,
        "imf": Namespace("http://ns.imfid.org/imf#"),
        "pav": Namespace("http://purl.org/pav/"),
        "foaf": FOAF._NS,
        "dexpi": Namespace("http://sandbox.dexpi.org/rdl/"),
        "qudt": Namespace("https://qudt.org/vocab/unit/"),
        "iodd": Namespace("http://www.io-link.com/IODD/2010/10/"),
        "aml": Namespace("https://www.automationml.org/"),
    }


DEFAULT_URI = ""

DEFAULT_DOCS_URL = "https://cognite-neat.readthedocs-hosted.com/en/latest/"

DMS_CONTAINER_PROPERTY_SIZE_LIMIT = 100
DMS_VIEW_CONTAINER_SIZE_LIMIT = 10
DMS_DIRECT_RELATION_LIST_LIMIT = 100

_ASSET_ROOT_PROPERTY = {
    "connection": "direct",
    "container": "cdf_cdm:CogniteAsset",
    "container_property": "assetHierarchy_root",
    "description": "An automatically updated reference to the top-level asset of the hierarchy.",
    "immutable": False,
    "is_list": False,
    "name": "Root",
    "nullable": True,
    "value_type": "cdf_cdm:CogniteAsset(version=v1)",
    "view": "cdf_cdm:CogniteAsset(version=v1)",
    "view_property": "root",
}

_ASSET_PATH_PROPERTY = {
    "connection": "direct",
    "container": "cdf_cdm:CogniteAsset",
    "container_property": "assetHierarchy_path",
    "description": (
        "An automatically updated ordered list of this asset's ancestors, starting with the root asset. "
        "Enables subtree filtering to find all assets under a parent."
    ),
    "immutable": False,
    "is_list": True,
    "name": "Path",
    "nullable": True,
    "value_type": "cdf_cdm:CogniteAsset(version=v1)",
    "view": "cdf_cdm:CogniteAsset(version=v1)",
    "view_property": "path",
}


def get_asset_read_only_properties_with_connection() -> "list[DMSProperty]":
    """Gets the asset read-only properties with connection, i.e. Root and Path."""
    from cognite.neat._rules.models.dms import DMSProperty

    return [DMSProperty.model_validate(item) for item in (_ASSET_ROOT_PROPERTY, _ASSET_PATH_PROPERTY)]


READONLY_PROPERTIES_BY_CONTAINER: Mapping[dm.ContainerId, frozenset[str]] = {
    dm.ContainerId("cdf_cdm", "CogniteAsset"): frozenset(
        {"assetHierarchy_root", "assetHierarchy_path", "assetHierarchy_path_last_updated_time"}
    ),
    dm.ContainerId("cdf_cdm", "CogniteFile"): frozenset({"isUploaded", "uploadedTime"}),
}

HIERARCHICAL_PROPERTIES_BY_CONTAINER: Mapping[dm.ContainerId, frozenset[str]] = {
    dm.ContainerId("cdf_cdm", "CogniteAsset"): frozenset({"assetHierarchy_parent"})
}


def is_readonly_property(container: dm.ContainerId, property_: str) -> bool:
    return container in READONLY_PROPERTIES_BY_CONTAINER and property_ in READONLY_PROPERTIES_BY_CONTAINER[container]


def is_hierarchy_property(container: dm.ContainerId, property_: str) -> bool:
    return (
        container in HIERARCHICAL_PROPERTIES_BY_CONTAINER
        and property_ in HIERARCHICAL_PROPERTIES_BY_CONTAINER[container]
    )


DMS_RESERVED_PROPERTIES = frozenset(
    {
        "createdTime",
        "deletedTime",
        "edge_id",
        "extensions",
        "externalId",
        "lastUpdatedTime",
        "node_id",
        "project_id",
        "property_group",
        "seq",
        "space",
        "version",
        "tg_table_name",
        "startNode",
        "endNode",
    }
)



from dataclasses import dataclass
from typing import Generic

from cognite.neat._issues._base import NeatError, ResourceType, T_Identifier, T_ReferenceIdentifier
from cognite.neat._utils.text import humanize_collection


@dataclass(unsafe_hash=True)
class ResourceError(NeatError, Generic[T_Identifier], RuntimeError):
    """Base class for resource errors {resource_type} with identifier {identifier}"""

    identifier: T_Identifier
    resource_type: ResourceType


@dataclass(unsafe_hash=True)
class ResourceDuplicatedError(ResourceError[T_Identifier]):
    """The {resource_type} with identifier {identifier} is duplicated in {location}"""

    fix = "Remove the duplicate {resource_type} {identifier}."
    location: str


@dataclass(unsafe_hash=True)
class ResourceRetrievalError(ResourceError[T_Identifier]):
    """Failed to retrieve {resource_type} with identifier {identifier}. The error was: {error}"""

    error: str


@dataclass(unsafe_hash=True)
class ResourceNotFoundError(ResourceError, Generic[T_Identifier, T_ReferenceIdentifier]):
    """The {resource_type} with identifier '{identifier}' does not exist"""

    extra = " This is expected by {referred_type} {referred_by}."

    fix = "Create the {resource_type} {identifier}"

    referred_by: T_ReferenceIdentifier | None = None
    referred_type: ResourceType | None = None
    more: str | None = None

    def as_message(self, include_type: bool = True) -> str:
        msg = (self.__doc__ or "").format(resource_type=self.resource_type, identifier=self.identifier)
        if self.referred_by and self.referred_type:
            msg += self.extra.format(referred_type=self.referred_type, referred_by=self.referred_by)
        if self.more:
            msg += f" {self.more}"
        msg += f" Fix {self.fix.format(resource_type=self.resource_type, identifier=self.identifier)}"
        return msg


@dataclass(unsafe_hash=True)
class ResourceNotDefinedError(ResourceError[T_Identifier]):
    """The {resource_type} {identifier} is not defined in the {location}"""

    extra = "{column_name} {row_number} in {sheet_name}"
    fix = "Define the {resource_type} {identifier} in {location}."

    location: str
    column_name: str | None = None
    row_number: int | None = None
    sheet_name: str | None = None


@dataclass(unsafe_hash=True)
class ResourceConversionError(ResourceError, ValueError):
    """Failed to convert the {resource_type} {identifier} to {target_format}: {reason}"""

    fix = "Check the error message and correct the rules."
    target_format: str
    reason: str


@dataclass(unsafe_hash=True)
class ResourceCreationError(ResourceError[T_Identifier], ValueError):
    """Failed to create {resource_type} with identifier {identifier}. The error was: {error}"""

    error: str


@dataclass(unsafe_hash=True)
class ResourceMissingIdentifierError(NeatError, ValueError):
    """The {resource_type} with name {name} is missing an identifier."""

    resource_type: ResourceType
    name: str | None = None


@dataclass(unsafe_hash=True)
class ResourceChangedError(ResourceError[T_Identifier]):
    """The {resource_type} with identifier {identifier} has changed{changed}"""

    fix = (
        "When extending model with extension set to addition or reshape, "
        "the {resource_type} properties must remain the same"
    )

    changed_properties: frozenset[str]
    changed_attributes: frozenset[str]

    def as_message(self, include_type: bool = True) -> str:
        if self.changed_properties:
            changed = f" properties {humanize_collection(self.changed_properties)}."
        elif self.changed_attributes:
            changed = f" attributes {humanize_collection(self.changed_attributes)}."
        else:
            changed = "."
        msg = (self.__doc__ or "").format(resource_type=self.resource_type, identifier=self.identifier, changed=changed)
        msg += f"Fix {self.fix.format(resource_type=self.resource_type)}"
        return msg



from pathlib import Path
from typing import Literal, cast

from rdflib import URIRef

from cognite.neat._client import NeatClient
from cognite.neat._graph.extractors import KnowledgeGraphExtractor
from cognite.neat._issues import IssueList
from cognite.neat._rules.importers import BaseImporter, InferenceImporter
from cognite.neat._rules.models import DMSRules, InformationRules
from cognite.neat._rules.transformers import (
    VerifiedRulesTransformer,
)
from cognite.neat._store import NeatGraphStore, NeatRulesStore
from cognite.neat._utils.upload import UploadResultList

from .exceptions import NeatSessionError, _session_method_wrapper


class SessionState:
    def __init__(
        self,
        store_type: Literal["memory", "oxigraph"],
        storage_path: Path | None = None,
        client: NeatClient | None = None,
    ) -> None:
        self.instances = InstancesState(store_type, storage_path=storage_path)
        self.rule_store = NeatRulesStore()
        self.last_reference: DMSRules | InformationRules | None = None
        self.client = client
        self.quoted_source_identifiers = False

    def rule_transform(self, *transformer: VerifiedRulesTransformer) -> IssueList:
        if not transformer:
            raise NeatSessionError("No transformers provided.")

        start = self.rule_store.provenance[-1].target_entity.display_name
        issues = self.rule_store.transform(*transformer)
        last_entity = self.rule_store.provenance[-1].target_entity
        issues.action = f"{start} &#8594; {last_entity.display_name}"
        issues.hint = "Use the .inspect.issues() for more details."
        return issues

    def rule_import(self, importer: BaseImporter, enable_manual_edit: bool = False) -> IssueList:
        issues = self.rule_store.import_rules(
            importer,
            client=self.client,
            enable_manual_edit=enable_manual_edit,
        )
        if self.rule_store.empty:
            result = "failed"
        else:
            result = self.rule_store.provenance[-1].target_entity.display_name
        if isinstance(importer, InferenceImporter):
            issues.action = f"Inferred {result}"
        else:
            issues.action = f"Read {result}"
        if issues:
            issues.hint = "Use the .inspect.issues() for more details."
        return issues

    def write_graph(self, extractor: KnowledgeGraphExtractor) -> IssueList:
        extract_issues = self.instances.store.write(extractor)
        issues = self.rule_store.import_graph(extractor)
        issues.extend(extract_issues)
        return issues


class InstancesState:
    def __init__(
        self,
        store_type: Literal["memory", "oxigraph"],
        storage_path: Path | None = None,
    ) -> None:
        self.store_type = store_type
        self.storage_path = storage_path
        self.issue_lists = IssueList()
        self.outcome = UploadResultList()
        # These contain prefixes added by Neat at the extraction stage.
        # We store them such that they can be removed in the load stage.
        self.neat_prefix_by_predicate_uri: dict[URIRef, str] = {}
        self.neat_prefix_by_type_uri: dict[URIRef, str] = {}

        # Ensure that error handling is done in the constructor
        self.store: NeatGraphStore = _session_method_wrapper(self._create_store, "NeatSession")()

        if self.storage_path:
            print("Remember to close neat session .close() once you are done to avoid oxigraph lock.")

    def _create_store(self) -> NeatGraphStore:
        if self.store_type == "oxigraph":
            if self.storage_path:
                self.storage_path.mkdir(parents=True, exist_ok=True)
            return NeatGraphStore.from_oxi_local_store(storage_dir=self.storage_path)
        else:
            return NeatGraphStore.from_memory_store()

    @property
    def empty(self) -> bool:
        return self.store.empty

    @property
    def last_outcome(self) -> UploadResultList:
        if not self.outcome:
            raise NeatSessionError(
                "No outcome available. Try using [bold].to.cdf.instances[/bold] to upload a data instance."
            )
        return cast(UploadResultList, self.outcome[-1])



from abc import ABC, abstractmethod
from collections.abc import Sequence

from cognite.neat._issues import NeatIssue


class Tracker(ABC):
    def __init__(self, name: str, units: list[str], unit_type: str) -> None:
        self.name = name
        self.units = units
        self.unit_type = unit_type

    @abstractmethod
    def start(self, unit: str) -> None:
        raise NotImplementedError()

    @abstractmethod
    def finish(self, unit: str) -> None:
        raise NotImplementedError()

    @abstractmethod
    def _issue(self, issue: NeatIssue) -> None:
        raise NotImplementedError()

    def issue(self, issue: NeatIssue | Sequence[NeatIssue]) -> None:
        if isinstance(issue, NeatIssue):
            self._issue(issue)
            return
        for item in issue:
            self._issue(item)



from ._base import BaseExporter, CDFExporter
from ._rules2dms import DMSExporter
from ._rules2excel import ExcelExporter
from ._rules2instance_template import InstanceTemplateExporter
from ._rules2ontology import GraphExporter, OWLExporter, SemanticDataModelExporter, SHACLExporter
from ._rules2yaml import YAMLExporter

__all__ = [
    "BaseExporter",
    "CDFExporter",
    "DMSExporter",
    "ExcelExporter",
    "GraphExporter",
    "InstanceTemplateExporter",
    "OWLExporter",
    "SHACLExporter",
    "SemanticDataModelExporter",
    "YAMLExporter",
]


def _repr_html_() -> str:
    import pandas as pd

    table = pd.DataFrame(  # type: ignore[operator]
        [
            {
                "Exporter": name,
                "Description": globals()[name].__doc__.strip().split("\n")[0] if globals()[name].__doc__ else "Missing",
            }
            for name in __all__
            if name not in ("BaseExporter", "CDFExporter", "GraphExporter")
        ]
    )._repr_html_()

    return (
        "<strong>Exporter</strong> An exporter converts Neat's representation of a data model called <em>Rules</em>"
        f" into a schema/data model for a target format.<br />{table}"
    )



from pathlib import Path
from typing import Any, cast

import yaml

from cognite.neat._issues import IssueList, MultiValueError, NeatIssue
from cognite.neat._issues.errors import (
    FileMissingRequiredFieldError,
    FileNotAFileError,
    FileNotFoundNeatError,
    FileTypeUnexpectedError,
)
from cognite.neat._issues.warnings import NeatValueWarning
from cognite.neat._rules._shared import ReadRules, T_InputRules
from cognite.neat._rules.models import INPUT_RULES_BY_ROLE, RoleTypes

from ._base import BaseImporter


class YAMLImporter(BaseImporter[T_InputRules]):
    """Imports the rules from a YAML file.

    Args:
        raw_data: The raw data to be imported.

    .. note::

        YAML files are typically used for storing rules when checked into version control systems, e.g., git-history.
        The advantage of using YAML files over Excel is that tools like git can show the differences between different
        versions of the rules.

    """

    def __init__(
        self,
        raw_data: dict[str, Any],
        read_issues: list[NeatIssue] | None = None,
        filepaths: list[Path] | None = None,
        source_name: str = "Unknown",
    ) -> None:
        self.raw_data = raw_data
        self._read_issues = IssueList(read_issues)
        self._filepaths = filepaths
        self._source_name = source_name

    @property
    def description(self) -> str:
        return f"YAML file {self._source_name} read as unverified data model"

    @classmethod
    def from_file(cls, filepath: Path, source_name: str = "Unknown") -> "YAMLImporter":
        if not filepath.exists():
            return cls({}, [FileNotFoundNeatError(filepath)])
        elif not filepath.is_file():
            return cls({}, [FileNotAFileError(filepath)])
        elif filepath.suffix not in [".yaml", ".yml"]:
            return cls({}, [FileTypeUnexpectedError(filepath, frozenset([".yaml", ".yml"]))])
        return cls(yaml.safe_load(filepath.read_text()), filepaths=[filepath], source_name=source_name)

    def to_rules(self) -> ReadRules[T_InputRules]:
        if self._read_issues.has_errors or not self.raw_data:
            self._read_issues.trigger_warnings()
            raise MultiValueError(self._read_issues.errors)

        issue_list = IssueList(title="YAML Importer", issues=self._read_issues)

        if not self._filepaths:
            issue_list.append(
                NeatValueWarning(
                    f"{type(self).__name__} was called without filepaths when there is content",
                )
            )
            metadata_file = Path()
        else:
            metadata_file_nullable = next((file for file in self._filepaths if file.stem == "metadata"), None)
            metadata_file = metadata_file_nullable or self._filepaths[0]

        if "metadata" not in self.raw_data:
            self._read_issues.append(FileMissingRequiredFieldError(metadata_file, "section", "metadata"))
            issue_list.trigger_warnings()
            raise MultiValueError(self._read_issues.errors)

        metadata = self.raw_data["metadata"]

        if "role" not in metadata:
            self._read_issues.append(FileMissingRequiredFieldError(metadata, "metadata", "role"))
            issue_list.trigger_warnings()
            raise MultiValueError(self._read_issues.errors)

        role_input = RoleTypes(metadata["role"])
        role_enum = RoleTypes(role_input)
        rules_cls = INPUT_RULES_BY_ROLE[role_enum]

        rules = cast(T_InputRules, rules_cls.load(self.raw_data))

        issue_list.trigger_warnings()
        if self._read_issues.has_errors:
            raise MultiValueError(self._read_issues.errors)

        return ReadRules[T_InputRules](rules, {})



from ._base import RulesTransformer, VerifiedRulesTransformer
from ._converters import (
    AddClassImplements,
    ChangeViewPrefix,
    ClassicPrepareCore,
    ConversionTransformer,
    ConvertToRules,
    DMSToInformation,
    DropModelViews,
    IncludeReferenced,
    InformationToDMS,
    MergeDMSRules,
    MergeInformationRules,
    PrefixEntities,
    SetIDDMSModel,
    StandardizeNaming,
    StandardizeSpaceAndVersion,
    SubsetDMSRules,
    SubsetInformationRules,
    ToCompliantEntities,
    ToDataProductModel,
    ToDMSCompliantEntities,
    ToEnterpriseModel,
    ToExtensionModel,
    ToSolutionModel,
)
from ._mapping import AsParentPropertyId, MapOneToOne, RuleMapper
from ._verification import VerifyAnyRules, VerifyDMSRules, VerifyInformationRules

__all__ = [
    "AddClassImplements",
    "AsParentPropertyId",
    "ChangeViewPrefix",
    "ClassicPrepareCore",
    "ConversionTransformer",
    "ConvertToRules",
    "DMSToInformation",
    "DropModelViews",
    "IncludeReferenced",
    "InformationToDMS",
    "MapOneToOne",
    "MergeDMSRules",
    "MergeInformationRules",
    "PrefixEntities",
    "RuleMapper",
    "RulesTransformer",
    "SetIDDMSModel",
    "StandardizeNaming",
    "StandardizeSpaceAndVersion",
    "SubsetDMSRules",
    "SubsetInformationRules",
    "ToCompliantEntities",
    "ToDMSCompliantEntities",
    "ToDataProductModel",
    "ToEnterpriseModel",
    "ToExtensionModel",
    "ToSolutionModel",
    "VerifiedRulesTransformer",
    "VerifyAnyRules",
    "VerifyDMSRules",
    "VerifyInformationRules",
]



from dataclasses import dataclass

from cognite.neat._issues import NeatError


@dataclass(unsafe_hash=True)
class NeatValueError(NeatError, ValueError):
    """{raw_message}"""

    raw_message: str


@dataclass(unsafe_hash=True)
class NeatTypeError(NeatError, TypeError):
    """{raw_message}"""

    raw_message: str


@dataclass(unsafe_hash=True)
class RegexViolationError(NeatError, ValueError):
    """The value '{value}' failed regex, {regex}, validation in {location}.
    Make sure that the name follows the regex pattern."""

    value: str
    regex: str
    location: str


@dataclass(unsafe_hash=True)
class NeatImportError(NeatError, ImportError):
    """The functionality requires {module}. You can include it
    in your neat installation with `pip install "cognite-neat[{neat_extra}]"`.
    """

    module: str
    neat_extra: str



from typing import cast
from urllib.parse import quote

from rdflib import Namespace, URIRef
from rdflib.query import ResultRow

from cognite.neat._utils.rdf_ import get_namespace, remove_namespace_from_uri

from ._base import BaseTransformerStandardised, RowTransformationOutput


class MakeConnectionOnExactMatch(BaseTransformerStandardised):
    description: str = "Adds property that contains id of reference to all references of given class in Rules"
    _use_only_once: bool = False
    _need_changes = frozenset({})

    def __init__(
        self,
        subject_type: URIRef,
        subject_predicate: URIRef,
        object_type: URIRef,
        object_predicate: URIRef,
        connection: URIRef | str | None = None,
        limit: int | None = None,
    ):
        self.subject_type = subject_type
        self.subject_predicate = subject_predicate
        self.object_type = object_type
        self.object_predicate = object_predicate
        subject_namespace = Namespace(get_namespace(subject_type))
        self.connection: URIRef = (
            subject_namespace[quote(connection.strip())]
            if isinstance(connection, str)
            else connection or subject_namespace[remove_namespace_from_uri(self.object_type).lower()]
        )

        self.limit = limit

    def _iterate_query(self) -> str:
        query = """SELECT DISTINCT ?subject ?object
                            WHERE {{
                                ?subject a <{subject_type}> ;
                                        <{subject_predicate}> ?value .
                                ?object a <{object_type}> ;
                                        <{object_predicate}> ?value .
                            }}"""

        if self.limit and isinstance(self.limit, int) and self.limit > 0:
            query += f" LIMIT {self.limit}"

        return query.format(
            subject_type=self.subject_type,
            subject_predicate=self.subject_predicate,
            object_type=self.object_type,
            object_predicate=self.object_predicate,
        )

    def _count_query(self) -> str:
        query = """SELECT (COUNT(DISTINCT ?subject) as ?count)
                    WHERE {{
                        ?subject a <{subject_type}> ;
                                <{subject_predicate}> ?value .
                        ?object a <{object_type}> ;
                                <{object_predicate}> ?value .
                    }}"""

        if self.limit and isinstance(self.limit, int) and self.limit > 0:
            query += f" LIMIT {self.limit}"

        return query.format(
            subject_type=self.subject_type,
            subject_predicate=self.subject_predicate,
            object_type=self.object_type,
            object_predicate=self.object_predicate,
        )

    def operation(self, query_result_row: ResultRow) -> RowTransformationOutput:
        row_output = RowTransformationOutput()

        subject, object = cast(tuple[URIRef, URIRef], query_result_row)

        row_output.add_triples.add((subject, self.connection, object))
        row_output.instances_modified_count += 1
        return row_output



import os
import subprocess
from contextlib import redirect_stderr, redirect_stdout, suppress
from dataclasses import dataclass, fields
from io import StringIO
from pathlib import Path
from typing import Literal, TypeAlias, get_args

from cognite.client import ClientConfig, CogniteClient
from cognite.client.config import global_config
from cognite.client.credentials import CredentialProvider, OAuthClientCredentials, OAuthInteractive, Token

from cognite.neat import _version
from cognite.neat._utils.auxiliary import local_import

__all__ = ["EnvironmentVariables", "get_cognite_client"]

_LOGIN_FLOW: TypeAlias = Literal["infer", "client_credentials", "interactive", "token"]
_VALID_LOGIN_FLOWS = get_args(_LOGIN_FLOW)
_CLIENT_NAME = f"CogniteNeat:{_version.__version__}"


def get_cognite_client(env_file_name: str) -> CogniteClient:
    """Instantiate a CogniteClient using environment variables. If the environment variables are not set, the user will
    be prompted to enter them.

    Args:
        env_file_name: The name of the .env file to look for in the repository root / current working directory. If
        the file is found, the variables will be loaded from the file. If the file is not found, the user will
        be prompted to enter the variables and the file will be created.

    Returns:
        CogniteClient: A CogniteClient instance.

    """
    if not env_file_name.endswith(".env"):
        raise ValueError("env_file_name must end with '.env'")
    global_config.disable_pypi_version_check = True
    # First try to load from .env file in repository root
    repo_root = _repo_root()
    if repo_root:
        with suppress(KeyError, FileNotFoundError, TypeError):
            variables = _from_dotenv(repo_root / env_file_name)
            client = variables.get_client()
            print(f"Found {env_file_name} file in repository root. Loaded variables from {env_file_name} file.")
            return client
    elif (Path.cwd() / env_file_name).exists():
        with suppress(KeyError, FileNotFoundError, TypeError):
            variables = _from_dotenv(Path.cwd() / env_file_name)
            client = variables.get_client()
            print(
                f"Found {env_file_name} file in current working directory. Loaded variables from {env_file_name} file."
            )
            return client
    # Then try to load from environment variables
    with suppress(KeyError):
        variables = EnvironmentVariables.create_from_environ()
        print("Loaded variables from environment variables.")
        return variables.get_client()
    # If not found, prompt the user
    variables = _prompt_user()
    if repo_root and _env_in_gitignore(repo_root, env_file_name):
        env_file = repo_root / env_file_name
        location = "repository root"
    elif repo_root:
        # We do not offer to create the file in the repository root if it is in .gitignore
        # as an inexperienced user might accidentally commit it.
        print("Cannot create .env file in repository root as there is no .env entry in the .gitignore.")
        return variables.get_client()
    else:
        env_file = Path.cwd() / env_file_name
        location = "current working directory"
    local_import("rich", "jupyter")
    from rich.prompt import Prompt

    answer = Prompt.ask(
        f"Do you store the variables in an {env_file_name} file in the {location} for easy reuse?",
        choices=["y", "n"],
    )
    if env_file.exists():
        answer = Prompt.ask(f"{env_file} already exists. Overwrite?", choices=["y", "n"])
    if answer == "y":
        env_file.write_text(variables.create_env_file())
        print(f"Created {env_file_name} file in {location}.")

    return variables.get_client()


@dataclass
class EnvironmentVariables:
    CDF_CLUSTER: str
    CDF_PROJECT: str
    LOGIN_FLOW: _LOGIN_FLOW = "infer"
    IDP_CLIENT_ID: str | None = None
    IDP_CLIENT_SECRET: str | None = None
    TOKEN: str | None = None

    IDP_TENANT_ID: str | None = None
    IDP_TOKEN_URL: str | None = None

    CDF_URL: str | None = None
    IDP_AUDIENCE: str | None = None
    IDP_SCOPES: str | None = None
    IDP_AUTHORITY_URL: str | None = None
    CDF_MAX_WORKERS: int | None = None
    CDF_TIMEOUT: int | None = None
    CDF_REDIRECT_PORT: int = 53_000

    def __post_init__(self):
        if self.LOGIN_FLOW.lower() not in _VALID_LOGIN_FLOWS:
            raise ValueError(f"LOGIN_FLOW must be one of {_VALID_LOGIN_FLOWS}")
        if self.IDP_TOKEN_URL and not self.IDP_TENANT_ID:
            prefix, suffix = "https://login.microsoftonline.com/", "/oauth2/v2.0/token"
            if self.IDP_TOKEN_URL.startswith(prefix) and self.IDP_TOKEN_URL.endswith(suffix):
                self.IDP_TENANT_ID = self.IDP_TOKEN_URL.removeprefix(prefix).removesuffix(suffix)

    @property
    def cdf_url(self) -> str:
        return self.CDF_URL or f"https://{self.CDF_CLUSTER}.cognitedata.com"

    @property
    def idp_token_url(self) -> str:
        if self.IDP_TOKEN_URL:
            return self.IDP_TOKEN_URL
        if not self.IDP_TENANT_ID:
            raise KeyError("IDP_TENANT_ID or IDP_TOKEN_URL must be set in the environment.")
        return f"https://login.microsoftonline.com/{self.IDP_TENANT_ID}/oauth2/v2.0/token"

    @property
    def idp_audience(self) -> str:
        return self.IDP_AUDIENCE or f"https://{self.CDF_CLUSTER}.cognitedata.com"

    @property
    def idp_scopes(self) -> list[str]:
        if self.IDP_SCOPES:
            return self.IDP_SCOPES.split(",")
        return [f"https://{self.CDF_CLUSTER}.cognitedata.com/.default"]

    @property
    def idp_authority_url(self) -> str:
        if self.IDP_AUTHORITY_URL:
            return self.IDP_AUTHORITY_URL
        if not self.IDP_TENANT_ID:
            raise KeyError("IDP_TENANT_ID or IDP_AUTHORITY_URL must be set in the environment.")
        return f"https://login.microsoftonline.com/{self.IDP_TENANT_ID}"

    @classmethod
    def create_from_environ(cls) -> "EnvironmentVariables":
        if "CDF_CLUSTER" not in os.environ or "CDF_PROJECT" not in os.environ:
            raise KeyError("CDF_CLUSTER and CDF_PROJECT must be set in the environment.", "CDF_CLUSTER", "CDF_PROJECT")

        return cls(
            CDF_CLUSTER=os.environ["CDF_CLUSTER"],
            CDF_PROJECT=os.environ["CDF_PROJECT"],
            LOGIN_FLOW=os.environ.get("LOGIN_FLOW", "infer"),  # type: ignore[arg-type]
            IDP_CLIENT_ID=os.environ.get("IDP_CLIENT_ID"),
            IDP_CLIENT_SECRET=os.environ.get("IDP_CLIENT_SECRET"),
            TOKEN=os.environ.get("TOKEN"),
            CDF_URL=os.environ.get("CDF_URL"),
            IDP_TOKEN_URL=os.environ.get("IDP_TOKEN_URL"),
            IDP_TENANT_ID=os.environ.get("IDP_TENANT_ID"),
            IDP_AUDIENCE=os.environ.get("IDP_AUDIENCE"),
            IDP_SCOPES=os.environ.get("IDP_SCOPES"),
            IDP_AUTHORITY_URL=os.environ.get("IDP_AUTHORITY_URL"),
            CDF_MAX_WORKERS=int(os.environ["CDF_MAX_WORKERS"]) if "CDF_MAX_WORKERS" in os.environ else None,
            CDF_TIMEOUT=int(os.environ["CDF_TIMEOUT"]) if "CDF_TIMEOUT" in os.environ else None,
            CDF_REDIRECT_PORT=int(os.environ.get("CDF_REDIRECT_PORT", 53_000)),
        )

    @classmethod
    def default(cls) -> "EnvironmentVariables":
        # This method is for backwards compatibility with the old config
        # It is not recommended to use this method.
        return cls(
            LOGIN_FLOW="client_credentials",
            CDF_CLUSTER="api",
            CDF_PROJECT="dev",
            IDP_TENANT_ID="common",
            IDP_CLIENT_ID="neat",
            IDP_CLIENT_SECRET="secret",
            IDP_SCOPES="project:read,project:write",
            CDF_TIMEOUT=60,
            CDF_MAX_WORKERS=3,
        )

    def get_credentials(self) -> CredentialProvider:
        method_by_flow = {
            "client_credentials": self.get_oauth_client_credentials,
            "interactive": self.get_oauth_interactive,
            "token": self.get_token,
        }
        if self.LOGIN_FLOW in method_by_flow:
            return method_by_flow[self.LOGIN_FLOW]()
        key_options: list[tuple[str, ...]] = []
        for method in method_by_flow.values():
            try:
                return method()
            except KeyError as e:
                key_options += e.args[1:]
        raise KeyError(
            f"LOGIN_FLOW={self.LOGIN_FLOW} requires one of the following environment set variables to be set.",
            *key_options,
        )

    def get_oauth_client_credentials(self) -> OAuthClientCredentials:
        if not self.IDP_CLIENT_ID or not self.IDP_CLIENT_SECRET:
            raise KeyError(
                "IDP_CLIENT_ID and IDP_CLIENT_SECRET must be set in the environment.",
                "IDP_CLIENT_ID",
                "IDP_CLIENT_SECRET",
            )
        return OAuthClientCredentials(
            client_id=self.IDP_CLIENT_ID,
            client_secret=self.IDP_CLIENT_SECRET,
            token_url=self.idp_token_url,
            audience=self.idp_audience,
            scopes=self.idp_scopes,
        )

    def get_oauth_interactive(self) -> OAuthInteractive:
        if not self.IDP_CLIENT_ID:
            raise KeyError("IDP_CLIENT_ID must be set in the environment.", "IDP_CLIENT_ID")
        return OAuthInteractive(
            client_id=self.IDP_CLIENT_ID,
            authority_url=self.idp_authority_url,
            redirect_port=self.CDF_REDIRECT_PORT,
            scopes=self.idp_scopes,
        )

    def get_token(self) -> Token:
        if not self.TOKEN:
            raise KeyError("TOKEN must be set in the environment", "TOKEN")
        return Token(self.TOKEN)

    def get_client(self) -> CogniteClient:
        config = ClientConfig(
            client_name=_CLIENT_NAME,
            project=self.CDF_PROJECT,
            credentials=self.get_credentials(),
            base_url=self.cdf_url,
            max_workers=self.CDF_MAX_WORKERS,
            timeout=self.CDF_TIMEOUT,
        )
        return CogniteClient(config)

    def create_env_file(self) -> str:
        lines: list[str] = []
        first_optional = True
        for field in fields(self):
            is_optional = hasattr(self, field.name.lower())
            if is_optional and first_optional:
                lines.append(
                    "# The below variables are the defaults, they are automatically constructed unless they are set."
                )
                first_optional = False
            name = field.name.lower() if is_optional else field.name
            value = getattr(self, name)
            if value is not None:
                if isinstance(value, list):
                    value = ",".join(value)
                lines.append(f"{field.name}={value}")
        return "\n".join(lines)


def _from_dotenv(evn_file: Path) -> EnvironmentVariables:
    if not evn_file.exists():
        raise FileNotFoundError(f"{evn_file} does not exist.")
    content = evn_file.read_text()
    valid_variables = {f.name for f in fields(EnvironmentVariables)}
    variables: dict[str, str] = {}
    for line in content.splitlines():
        if line.startswith("#") or "=" not in line:
            continue
        key, value = line.split("=", 1)
        if key in valid_variables:
            variables[key] = value
    return EnvironmentVariables(**variables)  # type: ignore[arg-type]


def _prompt_user() -> EnvironmentVariables:
    local_import("rich", "jupyter")
    from rich.prompt import Prompt

    try:
        variables = EnvironmentVariables.create_from_environ()
        continue_ = Prompt.ask(
            f"Use environment variables for CDF Cluster '{variables.CDF_CLUSTER}' "
            f"and Project '{variables.CDF_PROJECT}'? [y/n]",
            choices=["y", "n"],
            default="y",
        )
        if continue_ == "n":
            variables = _prompt_cluster_and_project()
    except KeyError:
        variables = _prompt_cluster_and_project()

    login_flow = Prompt.ask("Login flow", choices=[f for f in _VALID_LOGIN_FLOWS if f != "infer"])
    variables.LOGIN_FLOW = login_flow  # type: ignore[assignment]
    if login_flow == "token":
        token = Prompt.ask("Enter token")
        variables.TOKEN = token
        return variables

    variables.IDP_CLIENT_ID = Prompt.ask("Enter IDP Client ID")
    if login_flow == "client_credentials":
        variables.IDP_CLIENT_SECRET = Prompt.ask("Enter IDP Client Secret", password=True)
        tenant_id = Prompt.ask("Enter IDP_TENANT_ID (leave empty to enter IDP_TOKEN_URL instead)")
        if tenant_id:
            variables.IDP_TENANT_ID = tenant_id
        else:
            token_url = Prompt.ask("Enter IDP_TOKEN_URL")
            variables.IDP_TOKEN_URL = token_url
        optional = ["IDP_AUDIENCE", "IDP_SCOPES"]
    else:  # login_flow == "interactive"
        tenant_id = Prompt.ask("Enter IDP_TENANT_ID (leave empty to enter IDP_AUTHORITY_URL instead)")
        if tenant_id:
            variables.IDP_TENANT_ID = tenant_id
        else:
            variables.IDP_AUTHORITY_URL = Prompt.ask("Enter IDP_TOKEN_URL")
        optional = ["IDP_SCOPES"]

    defaults = "".join(f"\n - {name}: {getattr(variables, name.lower())}" for name in optional)
    use_defaults = Prompt.ask(
        f"Use default values for the following variables?{defaults}", choices=["y", "n"], default="y"
    )
    if use_defaults:
        return variables
    for name in optional:
        value = Prompt.ask(f"Enter {name}")
        setattr(variables, name, value)
    return variables


def _prompt_cluster_and_project() -> EnvironmentVariables:
    from rich.prompt import Prompt

    cluster = Prompt.ask("Enter CDF Cluster (example 'greenfield', 'bluefield', 'westeurope-1)")
    project = Prompt.ask("Enter CDF Project")
    return EnvironmentVariables(cluster, project)


def _repo_root() -> Path | None:
    # Redirecting stderr to suppress the error message if the command fails
    with suppress(Exception), redirect_stderr(StringIO()), redirect_stdout(StringIO()):
        result = subprocess.run("git rev-parse --show-toplevel".split(), stdout=subprocess.PIPE)
        if (output := result.stdout.decode().strip()) != "":
            return Path(output)
    return None


def _env_in_gitignore(repo_root: Path, env_file_name: str) -> bool:
    ignore_file = repo_root / ".gitignore"
    if not ignore_file.exists():
        return False
    else:
        ignored = {line.strip() for line in ignore_file.read_text().splitlines()}
        return env_file_name in ignored or "*.env" in ignored


if __name__ == "__main__":
    c = get_cognite_client(".env")
    print(c.iam.token.inspect())



"""This module performs importing of various formats to one of serializations for which
there are loaders to TransformationRules pydantic class."""

from cognite.neat._rules.importers._rdf._base import BaseRDFImporter
from cognite.neat._rules.importers._rdf._shared import parse_classes, parse_properties

CLASSES_QUERY = """
    SELECT ?class_ ?name ?description ?implements
    WHERE {{
        VALUES ?type {{ imf:BlockType imf:TerminalType imf:AttributeType }}
        ?class_ a ?type .

        OPTIONAL {{?class_ rdfs:subClassOf ?parent }}.
        OPTIONAL {{?class_ rdfs:label|skos:prefLabel ?name }}.
        OPTIONAL {{?class_ rdfs:comment|skos:definition ?description}}.


        # Add imf:Attribute as parent class when no parent is found
        BIND(IF(!bound(?parent) && ?type = imf:AttributeType, imf:Attribute, ?parent) AS ?implements)

        # FILTERS
        FILTER (!isBlank(?class_))
        FILTER (!bound(?implements) || !isBlank(?implements))

        FILTER (!bound(?name) || LANG(?name) = "" || LANGMATCHES(LANG(?name), "{language}"))
        FILTER (!bound(?description) || LANG(?description) = "" || LANGMATCHES(LANG(?description), "{language}"))
    }}
    """

PROPERTIES_QUERY = """
    SELECT ?class_ ?property_ ?name ?description ?value_type ?min_count ?max_count ?default
    WHERE
    {{
        # CASE 1: Handling Blocks and Terminals
        {{
            VALUES ?type {{ imf:BlockType imf:TerminalType }}
            ?class_ a ?type ;
                sh:property ?propertyShape .
                ?propertyShape sh:path ?property_ .

            OPTIONAL {{ ?property_ skos:prefLabel ?name . }}
            OPTIONAL {{ ?property_ skos:definition ?description . }}
            OPTIONAL {{ ?property_ rdfs:range ?range . }}

            OPTIONAL {{ ?propertyShape sh:minCount ?min_count . }}
            OPTIONAL {{ ?propertyShape sh:maxCount ?max_count . }}
            OPTIONAL {{ ?propertyShape sh:hasValue ?default . }}
            OPTIONAL {{ ?propertyShape sh:class | sh:qualifiedValueShape/sh:class ?valueShape . }}
        }}

        UNION

        # CASE 2: Handling Attributes
        {{
            ?class_ a imf:AttributeType .
            BIND(xsd:anyURI AS ?valueShape)
            BIND(imf:predicate AS ?property_)
            ?class_  ?property_ ?defaultURI .
            BIND(STR(?defaultURI) AS ?default)

        }}

        # Set the value type for the property based on sh:class, sh:qualifiedValueType or rdfs:range
        BIND(IF(BOUND(?valueShape), ?valueShape, IF(BOUND(?range) , ?range , ?valueShape)) AS ?value_type)

        FILTER (!isBlank(?property_))
        FILTER (!bound(?class_) || !isBlank(?class_))
        FILTER (!bound(?name) || LANG(?name) = "" || LANGMATCHES(LANG(?name), "{language}"))
        FILTER (!bound(?description) || LANG(?description) = "" || LANGMATCHES(LANG(?description), "{language}"))
    }}
    """


class IMFImporter(BaseRDFImporter):
    """Convert IMF Types provided as SHACL shapes to Input Rules."""

    @property
    def description(self) -> str:
        return f"IMF Types {self.source_name} read as unverified data model"

    def _to_rules_components(
        self,
    ) -> dict:
        classes, issue_list = parse_classes(self.graph, CLASSES_QUERY, self.language, self.issue_list)
        self.issue_list = issue_list
        properties, issue_list = parse_properties(self.graph, PROPERTIES_QUERY, self.language, self.issue_list)
        self.issue_list = issue_list

        components = {
            "Metadata": self._metadata,
            "Classes": list(classes.values()) if classes else [],
            "Properties": list(properties.values()) if properties else [],
        }

        return components



from abc import ABC, abstractmethod
from collections.abc import Hashable, Iterable
from pathlib import Path
from typing import ClassVar, Generic, TypeVar

from cognite.client import CogniteClient
from cognite.client.data_classes.capabilities import Capability

from cognite.neat._issues import IssueList, NeatIssue
from cognite.neat._issues.errors import AuthorizationError
from cognite.neat._store import NeatGraphStore
from cognite.neat._utils.auxiliary import class_html_doc
from cognite.neat._utils.upload import UploadResult, UploadResultList

T_Output = TypeVar("T_Output")


# Sentinel value to indicate in the load method that all instances of a class have been loaded.
# https://en.wikipedia.org/wiki/Sentinel_value
class _END_OF_CLASS: ...


class _START_OF_CLASS:
    def __init__(self, class_name: str | None = None):
        self.class_name = class_name


class BaseLoader(ABC, Generic[T_Output]):
    _new_line = "\n"
    _encoding = "utf-8"

    def __init__(self, graph_store: NeatGraphStore):
        self.graph_store = graph_store

    @abstractmethod
    def write_to_file(self, filepath: Path) -> None:
        raise NotImplementedError

    def load(self, stop_on_exception: bool = False) -> Iterable[T_Output | NeatIssue]:
        """Load the graph with data."""
        return (
            item  # type: ignore[misc]
            for item in self._load(stop_on_exception)
            if not (item is _END_OF_CLASS or isinstance(item, _START_OF_CLASS))
        )

    @abstractmethod
    def _load(
        self, stop_on_exception: bool = False
    ) -> Iterable[T_Output | NeatIssue | type[_END_OF_CLASS] | _START_OF_CLASS]:
        """Load the graph with data."""
        pass

    @classmethod
    def _repr_html_(cls) -> str:
        return class_html_doc(cls)


class CDFLoader(BaseLoader[T_Output]):
    _UPLOAD_BATCH_SIZE: ClassVar[int] = 1000

    def load_into_cdf(
        self, client: CogniteClient, dry_run: bool = False, check_client: bool = True
    ) -> UploadResultList:
        upload_result_by_name: dict[str, UploadResult] = {}
        for upload_result in self.load_into_cdf_iterable(client, dry_run, check_client):
            if last_result := upload_result_by_name.get(upload_result.name):
                upload_result_by_name[upload_result.name] = last_result.merge(upload_result)
            else:
                upload_result_by_name[upload_result.name] = upload_result

        return UploadResultList(upload_result_by_name.values())

    def load_into_cdf_iterable(
        self, client: CogniteClient, dry_run: bool = False, check_client: bool = True
    ) -> Iterable[UploadResult]:
        if check_client:
            missing_capabilities = client.iam.verify_capabilities(self._get_required_capabilities())
            if missing_capabilities:
                upload_result = UploadResult[Hashable](name=type(self).__name__)
                upload_result.issues.append(
                    AuthorizationError(action="Upload to CDF", reason=str(missing_capabilities))
                )
                yield upload_result
                return

        issues = IssueList()
        items: list[T_Output] = []
        last_class_name: str | None = None
        for result in self._load(stop_on_exception=False):
            if isinstance(result, NeatIssue):
                issues.append(result)
            elif result is _END_OF_CLASS:
                ...
            elif isinstance(result, _START_OF_CLASS):
                last_class_name = result.class_name
                continue
            else:
                # MyPy does not understand that 'else' means the item will be of type T_Output
                items.append(result)  # type: ignore[arg-type]

            if len(items) >= self._UPLOAD_BATCH_SIZE or result is _END_OF_CLASS:
                yield from self._upload_to_cdf(client, items, dry_run, issues, last_class_name)
                issues = IssueList()
                items = []
        if items:
            yield from self._upload_to_cdf(client, items, dry_run, issues, last_class_name)

    @abstractmethod
    def _get_required_capabilities(self) -> list[Capability]:
        raise NotImplementedError

    @abstractmethod
    def _upload_to_cdf(
        self,
        client: CogniteClient,
        items: list[T_Output],
        dry_run: bool,
        read_issues: IssueList,
        class_name: str | None = None,
    ) -> Iterable[UploadResult]:
        raise NotImplementedError
